class CLICommand:
    """
    Download data and pre-trained model including a generator, predictors for DAC
    """

    @staticmethod
    def add_arguments(parser):
        from mofreinforce.utils.download import DEFAULT_PATH
        add = parser.add_argument
        add ('target', nargs='+', help='download data and pretrained models including a generator and predictors for DAC')
        add ('--outdir', '-o', help=f'The Path where downloaded data will be stored. \n'
                                    f'default : (default) {DEFAULT_PATH} \n')
        add ('--remove_tarfile', '-r', action='store_true', help='remove tar.gz file for download database')

    @staticmethod
    def run(args):
        from mofreinforce.utils.download import (
            download_default,
        )

        func_dic = {'default': download_default,
                    }

        for stuff in args.target:
            if stuff not in func_dic.keys():
                raise ValueError(f'target must be {", ".join(func_dic.keys())}, not {stuff}')

        for stuff in args.target:
            func = func_dic[stuff]
            if func.__code__.co_argcount == 1:
                func(args.outdir)
            else:
                func(args.outdir, args.remove_tarfile)


import torch.nn as nn


class Pooler(nn.Module):
    def __init__(self, hidden_size, index=0):
        super().__init__()
        self.dense = nn.Linear(hidden_size, hidden_size)
        self.activation = nn.Tanh()
        self.index = index

    def forward(self, hidden_states):
        first_token_tensor = hidden_states[:, self.index]
        pooled_output = self.dense(first_token_tensor)
        pooled_output = self.activation(pooled_output)
        return pooled_output


class RegressionHead(nn.Module):
    """
    head for Regression
    """

    def __init__(self, hid_dim):
        super().__init__()
        self.fc = nn.Linear(hid_dim, 1)

    def forward(self, x):
        x = self.fc(x)
        return x


class ClassificationHead(nn.Module):
    """
    head for Classification
    """

    def __init__(self, hid_dim, n_classes):
        super().__init__()

        if n_classes == 2:
            self.fc = nn.Linear(hid_dim, 1)
            self.binary = True
        else:
            self.fc = nn.Linear(hid_dim, n_classes)
            self.binary = False

    def forward(self, x):
        x = self.fc(x)

        return x, self.binary


from sacred import Experiment

ex = Experiment("reinforce")


@ex.config
def config():
    seed = 0
    exp_name = ""
    gpu_idx = 0

    # load predictor
    predictor_load_path = ["model/predictor_qkh.ckpt"]
    mean = [None]
    std = [None]

    # get reward
    threshold = False
    """
    (1) threshold == True
     if |pred| >= |reward_max|, reward = 1, else : reward = |pred| / |reward_max|
    (2) threshold == False
    reward = |pred| / |reward_max|
    """
    reward_max = [1.]  # if you want to normalize by max reward.

    # load generator
    dataset_dir = "data/dataset_generator"
    generator_load_path = "model/generator.ckpt"

    # REINFORCE
    lr = 1e-4
    decay_ol = 1.
    decay_topo = 1.
    scheduler = "constant"  # warmup = get_linear_schedule_with_warmup, constant = get_constant_schedule
    early_stop = .5  # early_stop when the accuracy of scaffold is less than it.
    ratio_exploit = 0.  # ratio for exploitation (freeze)
    """ to decrease effect of topology when calculating loss.
    Given topology is the discrete feature, its loss could much larger than the sum of organic linker.
    It gives rise to training failure.
    """
    ratio_mask_mc = 0.  # ratio for masking mc of input_src

    # Trainer
    max_epochs = 20
    batch_size = 16
    accumulate_grad_batches = 2
    devices = 1
    num_nodes = 1
    precision = 16
    resume_from = None
    limit_train_batches = 500
    limit_val_batches = 30
    val_check_interval = 1.0
    test_only = False
    load_path = ""
    gradient_clip_val = None

    # tensorboard
    log_dir = "reinforce/logs"


@ex.named_config
def test():
    exp_name = "test"


"""
Q_kH
"""


@ex.named_config
def v0_scratch():
    exp_name = "v0_scratch"
    test_only = True

    # reward
    reward_max = [-50.]

    # predictor
    predictor_load_path = ["model/predictor_qkh.ckpt"]
    mean = [-19.408]
    std = [-9.172]


@ex.named_config
def v0_qkh():
    """
    omit mc in the input
    """
    exp_name = "v0_qkh"
    max_epochs = 20

    # reward
    reward_max = [-50.]

    # reinforce
    early_stop = 0.5  # early_stop when the accuracy of scaffold is less than it.
    ratio_exploit = .6  # ratio for exploitation
    ratio_mask_mc = .5  # ratio for masking mc of input_src

    # predictor
    predictor_load_path = ["model/predictor_qkh.ckpt"]
    mean = [-19.408]
    std = [-9.172]


"""
Selectivity (v1)
"""


@ex.named_config
def v1_scratch():
    exp_name = "v1_scratch"
    test_only = True

    # reward
    reward_max = [10.]

    # predictor
    predictor_load_path = ["model/predictor_selectivity.ckpt"]
    mean = [1.871]
    std = [1.922]


@ex.named_config
def v1_selectivity():
    """
    omit mc in the input
    """
    exp_name = "v1_selectivity"
    max_epochs = 20

    # reward
    reward_max = [10.]

    # reinforce
    early_stop = 0.5  # early_stop when the accuracy of scaffold is less than it.
    ratio_exploit = .6  # ratio for exploitation
    ratio_mask_mc = .5  # ratio for masking mc of input_src

    # predictor
    predictor_load_path = ["model/predictor_selectivity.ckpt"]
    mean = [1.871]
    std = [1.922]


"""
v2 multi objective
"""


@ex.named_config
def v2_multi_scratch():
    exp_name = "v2_multi_scratch"
    test_only = True
    log_dir = "reinforce/logs"

    # reward
    threshold = True
    reward_max = [-40., 1.]

    # predictor
    predictor_load_path = [
        "predictor/logs/regression_qkh_seed0_from_/version_0/checkpoints/best.ckpt",
        "model/predictor_selectivity.ckpt"
    ]
    mean = [-19.408, 1.871]
    std = [-9.172, 1.922]


@ex.named_config
def v2_multi():
    exp_name = "v2_multi"
    max_epochs = 20
    log_dir = "reinforce/logs"

    # reward
    threshold = True
    reward_max = [-40., 1.]

    # reinforce
    early_stop = 0.5  # early_stop when the accuracy of scaffold is less than it.
    ratio_exploit = .6  # ratio for exploitation
    ratio_mask_mc = .5  # ratio for masking mc of input_src

    # predictor
    predictor_load_path = [
        "predictor/logs/regression_qkh_seed0_from_/version_0/checkpoints/best.ckpt",
        "model/predictor_selectivity.ckpt"
    ]
    mean = [-19.408, 1.871]
    std = [-9.172, 1.922]


@ex.named_config
def v2_multi_discrete():
    exp_name = "v2_multi_discrete"
    max_epochs = 20
    log_dir = "reinforce/logs"

    # reward
    threshold = True
    reward_max = [-25., 1.]

    # reinforce
    early_stop = 0.  # early_stop when the accuracy of scaffold is less than it.
    ratio_exploit = .6  # ratio for exploitation
    ratio_mask_mc = .5  # ratio for masking mc of input_src

    # predictor
    predictor_load_path = [
        "predictor/logs/regression_qkh_seed0_from_/version_0/checkpoints/best.ckpt",
        "model/predictor_selectivity.ckpt"
    ]
    mean = [-19.408, 1.871]
    std = [-9.172, 1.922]


"""
Void Fraction (test)
"""


@ex.named_config
def vf_scratch():
    exp_name = "vf_scratch"
    test_only = True

    # load predictor
    predictor_load_path = ["predictor/logs/regression_vf_seed0_from_/version_0/checkpoints/best.ckpt"]


@ex.named_config
def v0_vf():
    """
    omit mc in the input
    """
    exp_name = "v0_vf"
    max_epochs = 20

    # load predictor
    predictor_load_path = ["predictor/logs/regression_vf_seed0_from_/version_0/checkpoints/best.ckpt"]

    # reinforce
    early_stop = 0.5  # early_stop when the accuracy of scaffold is less than it.
    ratio_exploit = .6  # ratio for exploitation
    ratio_mask_mc = .5  # ratio for masking mc of input_src



import torch
import json
from pathlib import Path
from torch.utils.data import Dataset


class MOFDataset(Dataset):
    def __init__(self,
                 dataset_dir,
                 split,
                 ):
        assert split in ["train", "test", "val"]

        # load dict_mof
        path_dict_mof = Path(dataset_dir, f"{split}.json")
        print(f"read file : {path_dict_mof}")
        self.dict_mof = json.load(open(path_dict_mof, "r"))
        self.mof_name = list(self.dict_mof.keys())


    def __len__(self):
        return len(self.mof_name)

    def __getitem__(self, idx):
        ret = dict()
        mof_name = self.mof_name[idx]

        ret.update(
            {
                "mof_name": mof_name,
            }
        )
        ret.update(self.dict_mof[mof_name])
        return ret

    @staticmethod
    def collate(batch, max_len):

        keys = set([key for b in batch for key in b.keys()])
        dict_batch = {k: [dic[k] if k in dic else None for dic in batch] for k in keys}
        # target
        dict_batch["target"] = torch.FloatTensor(dict_batch["target"])
        # mc (idx)
        dict_batch["mc"] = torch.LongTensor(dict_batch["mc"])
        # topo (idx)
        dict_batch["topo"] = torch.LongTensor(dict_batch["topo"])
        # ol (selfies)
        ol_len = max_len - 3 # cls, mc, topo
        dict_batch["ol"] = torch.LongTensor(
            [ol + [0] * (ol_len - len(ol)) if len(ol) < ol_len else ol[:ol_len] for ol in dict_batch["ol"]]
        )
        return dict_batch



import json
from sacred import Experiment

ex = Experiment("generator")

mc_to_idx = json.load(open("data/mc_to_idx.json"))
topo_to_idx = json.load(open("data/topo_to_idx.json"))

@ex.config
def config():
    seed = 0
    exp_name = "generator"
    log_dir = "generator/logs"
    loss_names = {"generator" : 1}

    # datamodule
    dataset_dir = "data/dataset_generator"
    batch_size = 256
    num_workers = 8 # recommend num_gpus * 4
    max_len = 128

    # transformer
    path_topo_to_idx = "data/topo_to_idx.json"
    path_mc_to_idx = "data/mc_to_idx.json"
    path_vocab = "data/vocab_to_idx.json"
    # input_dim = len(vocab_to_idx)
    # output_dim = len(vocab_to_idx)
    hid_dim = 256
    n_layers = 3
    n_heads = 8
    pf_dim = 512
    dropout = 0.1
    max_len = 128
    src_pad_idx = 0
    trg_pad_idx = 0

    # Trainer
    per_gpu_batchsize = 128
    num_nodes = 1
    num_devices = 2
    precision = 16
    resume_from = None
    val_check_interval = 1.0
    test_only = False
    load_path = ""
    gradient_clip_val = None

    # Optimizer Setting
    optim_type = "adam"  # adamw, adam, sgd (momentum=0.9)
    learning_rate = 5e-4
    weight_decay = 0
    decay_power = "constant"  # default polynomial decay, [cosine, constant, constant_with_warmup]
    max_epochs = 100
    max_steps = -1  # num_data * max_epoch // batch_size (accumulate_grad_batches)
    warmup_steps = 0.0  # int or float ( max_steps * warmup_steps)
    end_lr = 0
    lr_mult = 1  # multiply lr for downstream heads

@ex.named_config
def v0():
    exp_name = "v0"

@ex.named_config
def test():
    exp_name = "test"

@ex.named_config
def v0_test():
    exp_name = "v0_test"
    load_path = "model/generator.ckpt"

    test_only=True
    num_devices=1

"""
old experiments
"""
@ex.named_config
def v0_grad_clip():
    exp_name = "v0_grad_clip"

    gradient_clip_val = 0.5







import argparse
import textwrap
from importlib import import_module
from argparse import RawTextHelpFormatter
from mofreinforce import __version__

commands = [
    ('download', 'mofreinforce.cli.download'),
]


def main(prog='mofreinforce', version=__version__, commands=commands, args=None):
    parser = argparse.ArgumentParser(prog=prog, )
    parser.add_argument('--version', action='version',
                        version= '%(prog)s-{}'.format(version))
    parser.add_argument('-T', '--traceback', action='store_true')
    subparsers = parser.add_subparsers(title='Sub-command', dest='command')

    subparser = subparsers.add_parser('help',
                                      description='Help',
                                      help='Help for sub-command.')

    subparser.add_argument('helpcommand',
                           nargs='?',
                           metavar='sub-command',
                           help='Provide help for sub-command.')


    functions = {}
    parsers = {}
    for command, module_name in commands:
        cmd = import_module(module_name).CLICommand
        docstring = cmd.__doc__
        if docstring is None:
            # Backwards compatibility with GPAW
            short = cmd.short_description
            long = getattr(cmd, 'description', short)
        else:
            parts = docstring.split('\n', 1)
            if len(parts) == 1:
                short = docstring
                long = docstring
            else:
                short, body = parts
                long = short
                #long = short + '\n' + textwrap.dedent(body)
        subparser = subparsers.add_parser(
            command,
            formatter_class=RawTextHelpFormatter,
            help=short,
            description=long)
        cmd.add_arguments(subparser)
        functions[command] = cmd.run
        parsers[command] = subparser

    #if hook:
    #    args = hook(parser, args)
    #    args = hook(parser, args)
    #else:
    args = parser.parse_args(args)

    if args.command == 'help':
        if args.helpcommand is None:
            parser.print_help()
        else:
            parsers[args.helpcommand].print_help()
    elif args.command is None:
        parser.print_usage()
    else:
        f = functions[args.command]
        try:
            if f.__code__.co_argcount == 1:
                f(args)
            else:
                f(args, parsers[args.command])
        except KeyboardInterrupt:
            pass
        except Exception as x:
            if args.traceback:
                raise
            else:
                l1 = '{}: {}\n'.format(x.__class__.__name__, x)
                l2 = ('To get a full traceback, use: {} -T {} ...'
                      .format(prog, args.command))
                parser.error(l1 + l2)


import torch
from torch import nn

from .layers import EncoderLayer, DecoderLayer


class Transformer(nn.Module):
    """
    Trnasformer for MOF Generator
    """
    def __init__(self,
                 input_dim,
                 output_dim,
                 topo_dim,
                 mc_dim,
                 hid_dim,
                 n_layers,
                 n_heads,
                 pf_dim,
                 dropout,
                 max_len,
                 src_pad_idx,
                 trg_pad_idx,
                 ):
        super().__init__()

        self.encoder = Encoder(
            input_dim=input_dim,
            mc_dim=mc_dim,
            hid_dim=hid_dim,
            n_layers=n_layers,
            n_heads=n_heads,
            pf_dim=pf_dim,
            dropout=dropout,
            max_len=max_len,
        )
        self.decoder = Decoder(
            output_dim=output_dim,
            topo_dim=topo_dim,
            mc_dim=mc_dim,
            hid_dim=hid_dim,
            n_layers=n_layers,
            n_heads=n_heads,
            pf_dim=pf_dim,
            dropout=dropout,
            max_len=max_len,
        )
        self.src_pad_idx = src_pad_idx
        self.trg_pad_idx = trg_pad_idx

        self.apply(self.init_weights)

    def init_weights(self, m):
        if hasattr(m, 'weight') and m.weight.dim() > 1:
            nn.init.xavier_uniform_(m.weight.data)

    def make_src_mask(self, src):
        """
        make padding mask for src
        :param src: [B, src_len]
        :return: [B, 1, 1, src_len]
        """
        src_mask = (src != self.src_pad_idx).unsqueeze(1).unsqueeze(2)
        return src_mask

    def make_trg_mask(self, trg):
        """
        make padding and look-ahead mask for trg
        :param trg: [B, trg_len]
        :return: [B, 1, trg_len, trg_len]
        """
        trg_pad_mask = (trg != self.trg_pad_idx).unsqueeze(1).unsqueeze(2)  # [batch size, 1, 1, trg len]

        trg_len = trg.shape[1]
        trg_sub_mask = torch.tril(torch.ones((trg_len, trg_len))).bool()  # [trg len, trg len]

        trg_mask = trg_pad_mask & trg_sub_mask.to(trg_pad_mask.device)  # [batch size, 1, trg len, trg len]

        return trg_mask

    def forward(self, src, trg):
        """
        :param src: [B, src_len]
        :param trg: [B, trg_len]
        :return: [B, trg_len, vocab_dim] [B, n_heads, trg_len, src_len]
        """
        src_mask = self.make_src_mask(src)  # [batch size, 1, 1, src len]
        trg_mask = self.make_trg_mask(trg)  # [batch size, 1, trg len, trg len]

        # encoder
        enc_src = self.encoder(src, src_mask)  # [batch size, src len, hid dim]

        # decoder
        output = self.decoder(trg, enc_src, trg_mask, src_mask)
        # [batch size, trg len, output dim], [batch size, n heads, trg len, src len]
        return output


class Encoder(nn.Module):
    def __init__(self,
                 input_dim,
                 mc_dim,
                 hid_dim,
                 n_layers,
                 n_heads,
                 pf_dim,
                 dropout,
                 max_len,
                 max_conn=10):
        super().__init__()
        self.mc_embedding = nn.Embedding(mc_dim, hid_dim)
        self.num_embedding = nn.Embedding(max_conn, hid_dim) # num_conn of ol
        self.vocab_embedding = nn.Embedding(input_dim, hid_dim)
        self.pos_embedding = nn.Embedding(max_len, hid_dim)

        self.layers = nn.ModuleList([EncoderLayer(hid_dim,
                                                  n_heads,
                                                  pf_dim,
                                                  dropout,
                                                  )
                                     for _ in range(n_layers)])

        self.dropout = nn.Dropout(dropout)

        self.scale = torch.sqrt(torch.FloatTensor([hid_dim]))

    def forward(self, src, src_mask):
        """
        :param src: [B, src_len]
        :param src_mask: [B, 1, 1, src_len]
        :param num_conn: [B]
        :return: [B, src_len, hid_dim]
        """
        batch_size = src.shape[0]
        src_len = src.shape[1]

        pos = torch.arange(0, src_len).unsqueeze(0).repeat(batch_size, 1).to(src.device)  # [batch size, src len]

        src = torch.concat(
            [
                self.mc_embedding(src[:, 0].unsqueeze(-1)),
                self.num_embedding(src[:, 1].unsqueeze(-1)),
                self.vocab_embedding(src[:, 2:]),  # [batch size, src_len, hid dim]
            ],
            dim=1
        )

        src = self.dropout((src * self.scale.to(src.device)) + self.pos_embedding(pos))
        # [batch size, src len, hid dim]

        for layer in self.layers:
            src = layer(src, src_mask)  # [batch size, src len, hid dim]

        return src


class Decoder(nn.Module):
    def __init__(self,
                 output_dim,
                 topo_dim,
                 mc_dim,
                 hid_dim,
                 n_layers,
                 n_heads,
                 pf_dim,
                 dropout,
                 max_len,
                 ):
        super().__init__()

        self.topo_embedding = nn.Embedding(topo_dim, hid_dim)
        self.mc_embedding = nn.Embedding(mc_dim, hid_dim)
        self.vocab_embedding = nn.Embedding(output_dim, hid_dim)
        self.pos_embedding = nn.Embedding(max_len, hid_dim)

        self.layers = nn.ModuleList([DecoderLayer(hid_dim,
                                                  n_heads,
                                                  pf_dim,
                                                  dropout,
                                                  )
                                     for _ in range(n_layers)])

        self.fc_out_topo = nn.Linear(hid_dim, topo_dim)
        self.fc_out_mc = nn.Linear(hid_dim, mc_dim)
        self.fc_out_ol = nn.Linear(hid_dim, output_dim)

        self.dropout = nn.Dropout(dropout)

        self.scale = torch.sqrt(torch.FloatTensor([hid_dim]))

    def forward(self, trg, enc_src, trg_mask, src_mask):
        """
        if len(trg) == 1: topo_embedding
        elif len(trg) == 2: mc_embedding
        else: vocab_embedding
        :param trg: [B, trg_len]
        :param enc_src: [B, src_len, hid_dim]
        :param trg_mask: [B, 1, trg_len, trg_len]
        :param src_mask: [B, 1, 1, src_len]
        :return: [B, trg_len, vocab_dim] [B, n_heads, trg_len, src_len]
        """

        batch_size = trg.shape[0]
        trg_len = trg.shape[1]

        pos = torch.arange(0, trg_len).unsqueeze(0).repeat(batch_size, 1).to(trg.device)  # [batch size, trg len]

        if trg_len == 1: # [SOS]
            tok_emb = self.vocab_embedding(trg[:, 0].unsqueeze(-1))
        elif trg_len == 2: # [SOS, topo]
            tok_emb = torch.concat(
                    [
                        self.vocab_embedding(trg[:, 0].unsqueeze(-1)),
                        self.topo_embedding(trg[:, 1].unsqueeze(-1)),
                    ],
                    dim=1
                )
        elif trg_len == 3: # [SOS, topo, mc]
            tok_emb = torch.concat(
                [
                    self.vocab_embedding(trg[:, 0].unsqueeze(-1)),
                    self.topo_embedding(trg[:, 1].unsqueeze(-1)),
                    self.mc_embedding(trg[:, 2].unsqueeze(-1)),
                ],
                dim=1
            )
        else: # [SOS, topo, mc, ol]
            tok_emb = torch.concat(
                [
                    self.vocab_embedding(trg[:, 0].unsqueeze(-1)),
                    self.topo_embedding(trg[:, 1].unsqueeze(-1)),
                    self.mc_embedding(trg[:, 2].unsqueeze(-1)),
                    self.vocab_embedding(trg[:, 3:])
                ],
                dim=1
            )

        # [batch size, trg len, hid dim]
        trg = self.dropout((tok_emb * self.scale.to(trg.device)) + self.pos_embedding(pos))

        for layer in self.layers:
            trg, attention = layer(trg, enc_src, trg_mask, src_mask)
        # [batch size, trg len, hid dim], [batch size, n heads, trg len, src len]

        output = {}
        if trg_len == 1:
            output.update(
                {
                    "output_topo" : self.fc_out_topo(trg[:, 0]),  # [batch size, topo_dim]
                }
            )
        elif trg_len == 2:
            output.update(
                {
                    "output_topo" : self.fc_out_topo(trg[:, 0]),  # [batch size, topo_dim],
                    "output_mc" : self.fc_out_mc(trg[:, 1]),  # [batch size, mc_dim]
                }
            )
        else:
            output.update(
                {
                    "output_topo" : self.fc_out_topo(trg[:, 0]),  # [batch size, topo_dim],
                    "output_mc" : self.fc_out_mc(trg[:, 1]),  # [batch size, mc_dim]
                    "output_ol" : self.fc_out_ol(trg[:, 2:])  # [batch size, trg len, output dim]
                }
            )
        output.update({"attention" : attention})
        return output



import torch
import torch.nn as nn
from transformers import AdamW
from transformers import (
    get_polynomial_decay_schedule_with_warmup,
    get_cosine_schedule_with_warmup,
    get_constant_schedule,
    get_constant_schedule_with_warmup,
)
from utils.gadgets import Scalar, Accuracy


def init_weights(module):
    if isinstance(module, (nn.Linear, nn.Embedding)):
        module.weight.data.normal_(mean=0.0, std=0.02)
    elif isinstance(module, nn.LayerNorm):
        module.bias.data.zero_()
        module.weight.data.fill_(1.0)

    if isinstance(module, nn.Linear) and module.bias is not None:
        module.bias.data.zero_()


def set_metrics(pl_module):
    for split in ["train", "val"]:
        for k, v in pl_module.hparams.config["loss_names"].items():
            if v < 1:
                continue
            if k == "regression" or k == "vfr":
                setattr(pl_module, f"{split}_{k}_loss", Scalar())
                setattr(pl_module, f"{split}_{k}_mae", Scalar())
                setattr(pl_module, f"{split}_{k}_r2", Scalar())
            elif k == "generator":
                setattr(pl_module, f"{split}_{k}_loss", Scalar())
                setattr(pl_module, f"{split}_{k}_acc_topo", Scalar())
                setattr(pl_module, f"{split}_{k}_acc_mc", Scalar())
                setattr(pl_module, f"{split}_{k}_acc_ol", Scalar())
            else:
                setattr(pl_module, f"{split}_{k}_accuracy", Accuracy())
                setattr(pl_module, f"{split}_{k}_loss", Scalar())


def set_task(pl_module):
    pl_module.current_tasks = [
        k for k, v in pl_module.hparams.config["loss_names"].items() if v >= 1
    ]
    return


def epoch_wrapup(pl_module):
    phase = "train" if pl_module.training else "val"

    the_metric = 0
    for loss_name, v in pl_module.hparams.config["loss_names"].items():
        if v < 1:
            continue

        if loss_name == "regression" or loss_name == "vfr":
            # mse loss
            pl_module.log(
                f"{loss_name}/{phase}/loss_epoch",
                getattr(pl_module, f"{phase}_{loss_name}_loss").compute(),
            )
            getattr(pl_module, f"{phase}_{loss_name}_loss").reset()
            # mae loss
            value = getattr(pl_module, f"{phase}_{loss_name}_mae").compute()
            pl_module.log(
                f"{loss_name}/{phase}/mae_epoch",
                value,
            )
            getattr(pl_module, f"{phase}_{loss_name}_mae").reset()

            value = -value
        elif loss_name == "generator":
            acc_topo = getattr(pl_module, f"{phase}_{loss_name}_acc_topo").compute()
            pl_module.log(f"{loss_name}/{phase}/acc_topo", acc_topo)
            getattr(pl_module, f"{phase}_{loss_name}_acc_topo").reset()

            acc_mc = getattr(pl_module, f"{phase}_{loss_name}_acc_mc").compute()
            pl_module.log(f"{loss_name}/{phase}/acc_mc", acc_mc)
            getattr(pl_module, f"{phase}_{loss_name}_acc_mc").reset()

            acc_ol = getattr(pl_module, f"{phase}_{loss_name}_acc_ol").compute()
            pl_module.log(f"{loss_name}/{phase}/acc_ol", acc_ol)
            getattr(pl_module, f"{phase}_{loss_name}_acc_ol").reset()
            value = acc_topo + acc_mc + acc_ol

        else:
            value = getattr(pl_module, f"{phase}_{loss_name}_accuracy").compute()
            pl_module.log(f"{loss_name}/{phase}/accuracy_epoch", value)
            getattr(pl_module, f"{phase}_{loss_name}_accuracy").reset()
            pl_module.log(
                f"{loss_name}/{phase}/loss_epoch",
                getattr(pl_module, f"{phase}_{loss_name}_loss").compute(),
            )
            getattr(pl_module, f"{phase}_{loss_name}_loss").reset()

        the_metric += value

    pl_module.log(f"{phase}/the_metric", the_metric)


def set_schedule(pl_module):
    lr = pl_module.hparams.config["learning_rate"]
    wd = pl_module.hparams.config["weight_decay"]

    no_decay = [
        "bias",
        "LayerNorm.bias",
        "LayerNorm.weight",
        "norm.bias",
        "norm.weight",
        "norm1.bias",
        "norm1.weight",
        "norm2.bias",
        "norm2.weight",
    ]
    head_names = ["regression_head", "classification_head"]
    lr_mult = pl_module.hparams.config["lr_mult"]
    end_lr = pl_module.hparams.config["end_lr"]
    decay_power = pl_module.hparams.config["decay_power"]
    optim_type = pl_module.hparams.config["optim_type"]

    optimizer_grouped_parameters = [
        {
            "params": [
                p
                for n, p in pl_module.named_parameters()
                if not any(nd in n for nd in no_decay)  # not within no_decay
                   and not any(bb in n for bb in head_names)  # not within head_names
            ],
            "weight_decay": wd,
            "lr": lr,
        },
        {
            "params": [
                p
                for n, p in pl_module.named_parameters()
                if any(nd in n for nd in no_decay)  # within no_decay
                   and not any(bb in n for bb in head_names)  # not within head_names
            ],
            "weight_decay": 0.0,
            "lr": lr,
        },
        {
            "params": [
                p
                for n, p in pl_module.named_parameters()
                if not any(nd in n for nd in no_decay)  # not within no_decay
                   and any(bb in n for bb in head_names)  # within head_names
            ],
            "weight_decay": wd,
            "lr": lr * lr_mult,
        },
        {
            "params": [
                p
                for n, p in pl_module.named_parameters()
                if any(nd in n for nd in no_decay) and any(bb in n for bb in head_names)
                # within no_decay and head_names
            ],
            "weight_decay": 0.0,
            "lr": lr * lr_mult,
        },
    ]

    if optim_type == "adamw":
        optimizer = AdamW(optimizer_grouped_parameters, lr=lr, eps=1e-8, betas=(0.9, 0.98)
                          )
    elif optim_type == "adam":
        optimizer = torch.optim.Adam(optimizer_grouped_parameters, lr=lr)
    elif optim_type == "sgd":
        optimizer = torch.optim.SGD(optimizer_grouped_parameters, lr=lr, momentum=0.9)

    if pl_module.trainer.max_steps == -1:
        max_steps = (
                len(pl_module.trainer.datamodule.train_dataloader())
                * pl_module.trainer.max_epochs
                // pl_module.trainer.accumulate_grad_batches
                // (pl_module.trainer.num_devices * pl_module.trainer.num_nodes)
        )
    else:
        max_steps = pl_module.trainer.max_steps

    warmup_steps = pl_module.hparams.config["warmup_steps"]
    if isinstance(pl_module.hparams.config["warmup_steps"], float):
        warmup_steps = int(max_steps * warmup_steps)

    print(f"max_epochs: {pl_module.trainer.max_epochs} | max_steps: {max_steps} | warmup_steps : {warmup_steps} "
          f"| weight_decay : {wd} | decay_power : {decay_power}")

    if decay_power == "cosine":
        scheduler = get_cosine_schedule_with_warmup(
            optimizer,
            num_warmup_steps=warmup_steps,
            num_training_steps=max_steps,
        )
    elif decay_power == 'constant':
        scheduler = get_constant_schedule(
            optimizer,
        )
    elif decay_power == 'constant_with_warmup':
        scheduler = get_constant_schedule_with_warmup(
            optimizer,
            num_warmup_steps=warmup_steps,
        )
    else:
        scheduler = get_polynomial_decay_schedule_with_warmup(
            optimizer,
            num_warmup_steps=warmup_steps,
            num_training_steps=max_steps,
            lr_end=end_lr,
            power=decay_power,
        )

    sched = {"scheduler": scheduler, "interval": "step"}

    return (
        [optimizer],
        [sched],
    )


class Normalizer(object):
    """
    normalize for regression
    """

    def __init__(self, mean, std):
        if mean and std:
            self._norm_func = lambda tensor: (tensor - mean) / std
            self._denorm_func = lambda tensor: tensor * std + mean
        else:
            self._norm_func = lambda tensor: tensor
            self._denorm_func = lambda tensor: tensor

        self.mean = mean
        self.std = std

    def encode(self, tensor):
        return self._norm_func(tensor)

    def decode(self, tensor):
        return self._denorm_func(tensor)



import os
import wget
import tarfile
from pathlib import Path
from mofreinforce import __root_dir__


DEFAULT_PATH = Path(__root_dir__)

class DownloadError(Exception):
    pass


def _remove_tmp_file(direc:Path):
    tmp_list = direc.parent.glob('*.tmp')
    for tmp in tmp_list:
        if tmp.exists():
            os.remove(tmp)


def _download_file(link, direc, name='target'):
    if direc.exists():
        print (f'{name} already exists.')
        return
    try:
        print(f'\n====Download {name} =============================================\n')
        filename = wget.download(link, out=str(direc))
    except KeyboardInterrupt:
        _remove_tmp_file(direc)
        raise
    except Exception as e:
        _remove_tmp_file(direc)
        raise DownloadError(e)
    else:
        print (f'\n====Successfully download : {filename}=======================================\n')


def download_default(direc=None, remove_tarfile=False):
    """
    downlaod data and pre-trained models including a generator, predictors for DAC
    """
    if not direc:
        direc = Path(DEFAULT_PATH)
        if not direc.exists():
            direc.mkdir(parents=True, exist_ok=True)
        direc = direc/'default.tar.gz'
    else:
        direc = Path(direc)
        if direc.is_dir():
            if not direc.exists():
                direc.mkdir(parents=True, exist_ok=True)
            direc = direc / 'default.tar.gz'
        else:
            raise ValueError(f'direc must be path for directory, not {direc}')

    link = 'https://figshare.com/ndownloader/files/39229598'
    name = 'basic data and pretrained models'
    _download_file(link, direc, name)

    print(f'\n====Unzip : {name}===============================================\n')
    with tarfile.open(direc) as f:
        f.extractall(path=direc.parent)

    print(f'\n====Unzip successfully: {name}===============================================\n')

    if remove_tarfile:
        os.remove(direc)



from functools import partial

import torch
import torch.nn as nn
from timm.models.layers import DropPath, trunc_normal_

class Mlp(nn.Module):
    def __init__(
            self,
            in_features,
            hidden_features=None,
            out_features=None,
            act_layer=nn.GELU,
            drop=0.0,
    ):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x


class Attention(nn.Module):
    def __init__(
            self,
            dim,
            num_heads=8,
            qkv_bias=False,
            qk_scale=None,
            attn_drop=0.0,
            proj_drop=0.0,
    ):
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads
        # NOTE scale factor was wrong in my original version, can set manually to be compat with prev weights
        self.scale = qk_scale or head_dim ** -0.5

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x, mask=None):
        B, N, C = x.shape
        assert C % self.num_heads == 0
        qkv = (
            self.qkv(x)  # [B, N, 3*C]
                .reshape(B, N, 3, self.num_heads, C // self.num_heads)  # [B, N, 3, num_heads, C//num_heads]
                .permute(2, 0, 3, 1, 4)  # [3, B, num_heads, N, C//num_heads]
        )
        q, k, v = (
            qkv[0],  # [B, num_heads, N, C//num_heads]
            qkv[1],  # [B, num_heads, N, C//num_heads]
            qkv[2],  # [B, num_heads, N, C//num_heads]
        )  # make torchscript happy (cannot use tensor as tuple)

        attn = (q @ k.transpose(-2, -1)) * self.scale  # [B, num_heads, N, N]
        if mask is not None:
            mask = mask.bool()
            attn = attn.masked_fill(~mask[:, None, None, :], float("-inf"))
        attn = attn.softmax(dim=-1)  # [B, num_heads, N, N]
        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B, N, C)  # [B, num_heads, N, C//num_heads] -> [B, N, C]
        x = self.proj(x)
        x = self.proj_drop(x)
        return x, attn


class Block(nn.Module):
    def __init__(
            self,
            dim,
            num_heads,
            mlp_ratio=4.0,
            qkv_bias=False,
            qk_scale=None,
            drop=0.0,
            attn_drop=0.0,
            drop_path=0.0,
            act_layer=nn.GELU,
            norm_layer=nn.LayerNorm,
    ):
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.attn = Attention(
            dim,
            num_heads=num_heads,
            qkv_bias=qkv_bias,
            qk_scale=qk_scale,
            attn_drop=attn_drop,
            proj_drop=drop,
        )
        # NOTE: drop path for stochastic depth, we shall see if this is better than dropout here
        self.drop_path = DropPath(drop_path) if drop_path > 0.0 else nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(
            in_features=dim,
            hidden_features=mlp_hidden_dim,
            act_layer=act_layer,
            drop=drop,
        )

    def forward(self, x, mask=None):
        _x, attn = self.attn(self.norm1(x), mask=mask)
        x = x + self.drop_path(_x)
        x = x + self.drop_path(self.mlp(self.norm2(x)))
        return x, attn


class Transformer(nn.Module):
    def __init__(self, embed_dim,
                 depth=12,
                 num_heads=12,
                 mlp_ratio=4.0,
                 qkv_bias=True,
                 qk_scale=None,
                 drop_rate=0.0,
                 attn_drop_rate=0.0,
                 drop_path_rate=0.0,
                 norm_layer=None,
                 add_norm_before_transformer=False,
                 config=None):
        super().__init__()


        norm_layer = norm_layer or partial(nn.LayerNorm, eps=1e-6)
        self.add_norm_before_transformer = add_norm_before_transformer
        if add_norm_before_transformer:
            self.pre_norm = norm_layer(embed_dim)

        dpr = [
            x.item() for x in torch.linspace(0, drop_path_rate, depth)
        ]  # stochastic depth decay rule
        self.blocks = nn.ModuleList(
            [
                Block(
                    dim=embed_dim,
                    num_heads=num_heads,
                    mlp_ratio=mlp_ratio,
                    qkv_bias=qkv_bias,
                    qk_scale=qk_scale,
                    drop=drop_rate,
                    attn_drop=attn_drop_rate,
                    drop_path=dpr[i],
                    norm_layer=norm_layer,
                )
                for i in range(depth)
            ]
        )
        self.norm = norm_layer(embed_dim)

        # trunc_normal_(self.cls_token, std=0.02)
        self.pos_drop = nn.Dropout(p=drop_rate)

        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=0.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)






import torch.nn.functional as F
from torchmetrics.functional import accuracy


def compute_loss(pl_module, batch):
    infer = pl_module.infer(batch)

    tgt_label = infer["tgt_label"]  # [B, seq_len]

    batch_size, _, vocab_dim = infer["output_ol"].shape

    # loss topo
    logit_topo = infer["output_topo"]
    label_topo = tgt_label[:, 0]
    loss_topo = F.cross_entropy(logit_topo, label_topo)
    # loss mc
    logit_mc = infer["output_mc"]
    label_mc = tgt_label[:, 1]
    loss_mc = F.cross_entropy(logit_mc, label_mc)
    # loss ol
    logit_ol = infer["output_ol"].reshape(-1, vocab_dim)
    label_ol = tgt_label[:, 2:].reshape(-1)
    loss_ol = F.cross_entropy(logit_ol, label_ol, ignore_index=0)
    # total loss
    total_loss = loss_topo + loss_mc + loss_ol

    ret = {
        "gen_loss": total_loss,
        "gen_labels": tgt_label,
    }

    # call update() loss and acc
    loss_name = "generator"
    phase = "train" if pl_module.training else "val"
    total_loss = getattr(pl_module, f"{phase}_{loss_name}_loss")(ret["gen_loss"])

    # acc
    acc_topo = getattr(pl_module, f"{phase}_{loss_name}_acc_topo")(
        accuracy(logit_topo.argmax(-1), label_topo)
    )
    acc_mc = getattr(pl_module, f"{phase}_{loss_name}_acc_mc")(
        accuracy(logit_mc.argmax(-1), label_mc)
    )
    acc_ol = getattr(pl_module, f"{phase}_{loss_name}_acc_ol")(
        accuracy(logit_ol.argmax(-1), label_ol, ignore_index=0)
    )

    pl_module.log(f"{loss_name}/{phase}/total_loss", total_loss, batch_size=batch_size, prog_bar=True, sync_dist=True)
    pl_module.log(f"{loss_name}/{phase}/loss_topo", loss_topo, batch_size=batch_size, prog_bar=False, sync_dist=True)
    pl_module.log(f"{loss_name}/{phase}/loss_mc", loss_mc, batch_size=batch_size, prog_bar=False, sync_dist=True)
    pl_module.log(f"{loss_name}/{phase}/loss_ol", loss_ol, batch_size=batch_size, prog_bar=False, sync_dist=True)
    pl_module.log(f"{loss_name}/{phase}/acc_topo", acc_topo, batch_size=batch_size, prog_bar=True, sync_dist=True)
    pl_module.log(f"{loss_name}/{phase}/acc_mc", acc_mc, batch_size=batch_size, prog_bar=True, sync_dist=True)
    pl_module.log(f"{loss_name}/{phase}/acc_ol", acc_ol, batch_size=batch_size, prog_bar=True, sync_dist=True)
    return ret



import os
from functools import partial

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.distributions import Categorical
from torch.utils.tensorboard.writer import SummaryWriter

from rdkit import Chem
from rdkit.Chem.Draw import MolToImage

import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np
import libs.selfies as sf

class Reinforce(object):
    def __init__(self,
                 generator,
                 rmsd_predictor,
                 target_predictor,
                 get_reward_rmsd,
                 get_reward_target,
                 vocab_to_idx,
                 mc_to_idx,
                 topo_to_idx,
                 emb_dim=128,
                 hid_dim=128,
                 config=None,
                 ):
        """
        REINFORCE algorithm to generate MOFs to maximize reward functions (rsmsd and target)

        Parameters;
        generator_v0 (nn.Module): generator_v0
        rmsd_predictor (nn.Module): predictor for rmsd
        target_predictor (nn.Module): predictor for target
        get_reward_rmsd (function): get rewards for rmsd
        get_rewad_target (function): get rewards for target
        vocab_to_idx (dictionary) dictionary for token of selfies with index
        mc_to_idx (dictionary): dictionary for metal cluster with index
        topo_to_idx (dictionary): dictionary for topology with index
        emb_dim (int): dimension of embedding for metal cluster and topology
        hid_dim (int): dimension of hidden for metal culster and topology
        """
        super(Reinforce, self).__init__()
        self.generator = generator
        self.rmsd_predictor = rmsd_predictor
        self.target_predictor = target_predictor
        self.get_reward_rmsd = partial(get_reward_rmsd,
                                       criterion=config["criterion_rmsd"],
                                       reward_positive=config["reward_positive_rmsd"],
                                       reward_negative=config["reward_negative_rmsd"],
                                       )
        self.get_reward_target = partial(get_reward_target,
                                         criterion=config["criterion_target"],
                                         reward_positive=config["reward_positive_target"],
                                         reward_negative=config["reward_negative_target"],
                                         )

        self.vocab_to_idx = vocab_to_idx
        self.mc_to_idx = mc_to_idx
        self.topo_to_idx = topo_to_idx
        self.emb_dim = emb_dim
        self.hid_dim = hid_dim

        self.rmsd_predictor.eval()
        self.target_predictor.eval()
        self.upgrade_generator()

        self.log_dir = f"{config['log_dir']}/{config['exp_name']}"
        if os.path.exists(self.log_dir) and config["test_only"] is not True:
            raise Exception(f"{self.log_dir} is already exist")

        self.writer = SummaryWriter(log_dir=self.log_dir)

    def upgrade_generator(self):
        """
        add metal clutser and topology to generator_v0
        """
        self.generator.embedding_topo = nn.Embedding(len(self.topo_to_idx), self.emb_dim)
        self.generator.rc_mc_1 = nn.Linear(self.emb_dim, self.hid_dim)
        self.generator.rc_mc_2 = nn.Linear(self.hid_dim, len(self.mc_to_idx))

        self.generator.optimizer = self.generator.optimizer_instance(self.generator.parameters(),
                                                                     lr=1e-3,
                                                                     weight_decay=0.00001)
        self.generator = self.generator.cuda()

    def load_model(self, load_filename):
        self.generator.load_state_dict(torch.load(load_filename))

    def generate_mc(self, topo, return_loss=False):
        emb_topo = self.generator.embedding_topo(topo)
        logits = F.relu(self.generator.rc_mc_1(emb_topo))
        pred_mc = F.softmax(self.generator.rc_mc_2(logits), dim=-1)
        m = Categorical(pred_mc)
        mc = m.sample()
        loss_mc = torch.log(pred_mc[mc])

        if return_loss:
            return mc, loss_mc
        else:
            return mc

    def generate_ol(self):
        num_fail = 0
        # generate smiles
        valid_smiles = False
        while not valid_smiles:
            gen_sf = self.generator.evaluate()
            try:
                gen_sm = sf.decoder(gen_sf[5:-5])
                m = Chem.MolFromSmiles(gen_sm)
            except Exception:
                num_fail += 1
                continue

            # encodig_sf
            encoded_sf = sf.selfies_to_encoding(selfies=gen_sf,
                                                vocab_stoi=self.vocab_to_idx,
                                                enc_type="label")

            if m and 10 < len(encoded_sf) <= 100:
                valid_smiles = True
            else:
                num_fail += 1

        return encoded_sf, gen_sm, num_fail

    def init_generator(self):
        """
        Initialize stackRNN
        """

        hidden = self.generator.init_hidden()
        if self.generator.has_cell:
            cell = self.generator.init_cell()
            hidden = (hidden, cell)
        if self.generator.has_stack:
            stack = self.generator.init_stack()
        else:
            stack = None

        return hidden, stack

    def policy_gradient(self, n_batch, gamma, topo_idx=-1):
        """
        REINFORCE algorithm
        """
        self.generator.train()

        rl_loss = 0
        self.generator.optimizer.zero_grad()
        total_reward = 0
        num_fail = 0

        for n_epi in range(n_batch):
            # topology
            if topo_idx < 0:
                topo = torch.randint(0, len(self.topo_to_idx), size=(1,))[0].cuda()
            else:
                topo = torch.tensor(topo_idx).cuda()

            # metal cluster
            mc, loss_mc = self.generate_mc(topo, return_loss=True)
            rl_loss -= loss_mc

            # organic linker
            encoded_sf, _, n_f = self.generate_ol()
            num_fail += n_f

            # rmsd reward
            reward_rmsd, output_rmsd = self.get_reward_rmsd(topo, mc, encoded_sf, self.rmsd_predictor)

            # target reward
            reward_target, output_target = self.get_reward_target(topo, mc, encoded_sf, self.target_predictor)

            # REINFORCE algorithm
            discounted_reward = reward_rmsd + reward_target
            total_reward += reward_rmsd + reward_target

            # accumulate trajectory
            hidden, stack = self.init_generator()

            #  accumulate trajectory
            trajectory = torch.LongTensor(encoded_sf).cuda()

            for p in range(len(trajectory) - 1):

                output, hidden, stack = self.generator(trajectory[p], hidden, stack)
                log_probs = F.log_softmax(output, dim=-1)
                top_i = trajectory[p + 1]
                rl_loss -= (log_probs[0, top_i] * discounted_reward)
                discounted_reward = discounted_reward * gamma

        # backpropagation
        rl_loss = rl_loss / n_batch
        total_reward = total_reward / n_batch
        num_fail = num_fail / n_batch

        rl_loss.backward()
        self.generator.optimizer.step()

        return total_reward, rl_loss.item(), num_fail

    def test_estimate(self, n_to_generate, topo_idx=-1):
        with torch.no_grad():
            self.generator.evaluate()
            gen_mofs = []
            rewards = {"rmsd": [], "target": []}
            outputs = {"rmsd": [], "target": []}
            num_fail = 0

            for i in range(n_to_generate):
                # topology
                if topo_idx < 0:
                    topo = torch.randint(0, len(self.topo_to_idx), size=(1,))[0].cuda()
                else:
                    topo = torch.tensor(topo_idx).cuda()

                # metal cluster
                mc = self.generate_mc(topo)

                # organic linker
                encoded_sf, gen_sm, n_f = self.generate_ol()
                num_fail += n_f
                gen_mofs.append([topo.detach().item(), mc.detach().item(), gen_sm])

                # rmsd reward
                reward_rmsd, output_rmsd = self.get_reward_rmsd(topo, mc, encoded_sf, self.rmsd_predictor)

                # target reward
                reward_target, output_target = self.get_reward_target(topo, mc, encoded_sf, self.target_predictor)

                rewards["rmsd"].append(reward_rmsd)
                rewards["target"].append(reward_target)

                outputs["rmsd"].append(output_rmsd)
                outputs["target"].append(output_target)

        return rewards, outputs, num_fail, gen_mofs


    def write_logs_test_and_estimate(self, test_reward, test_output, test_num_fail, gen_mofs, n_iter, n_to_generate):
        reward_rmsd = np.array(test_reward["rmsd"])
        reward_target = np.array(test_reward["target"])
        output_rmsd = np.array(test_output["rmsd"])
        output_target = np.array(test_output["target"])

        # add scalar to log
        self.writer.add_scalar("test/reward/rmsd", reward_rmsd.mean(), n_iter)
        self.writer.add_scalar("test/reward/target", reward_target.mean(), n_iter)
        self.writer.add_scalar("test/output/rmsd", output_rmsd.mean(), n_iter)
        self.writer.add_scalar("test/output/target", output_target.mean(), n_iter)
        self.writer.add_scalar("test/num_fail", test_num_fail / n_to_generate, n_iter)
        # add histogram to log
        fig, axes = plt.subplots(2, 2, figsize=(12, 8))
        data = [reward_rmsd, reward_target, output_rmsd, output_target]
        title = ["reward_rmsd", "reward_target", "output_rmsd", "output_target"]
        for i in range(4):
            ax = axes[i // 2, i % 2]
            ax.set_title(title[i])
            sns.histplot(data[i], ax=ax)
        self.writer.add_figure("test", fig, n_iter)

        # add image to log
        imgs = []
        for i, m in enumerate(gen_mofs):
            m = Chem.MolFromSmiles(m[2])
            img = MolToImage(m)
            img = np.array(img)
            img = torch.tensor(img)
            imgs.append(img)
        imgs = np.stack(imgs, axis=0)
        self.writer.add_images("test/gen_ol/", imgs, n_iter, dataformats="NHWC")


    def train(self,
              n_iters=10000,
              n_print=100,
              n_to_generate=200,
              n_batch=10,
              gamma=0.80,
              topo_idx=-1,
              ):
        """

        :param n_iters:  number of iterations
        :param n_print: number of iterations to print
        :param n_to_generate: how many mof will be generated when test_and_estimate
        :param n_batch: number of batch size
        :param gamma: discount ratio of REINFORCE
        :param topo_idx: if topo_idx < 0, randomly selecting topology
        :return: None
        """

        reward = 0
        loss = 0
        num_fail = 0
        callback = 0

        for n_iter in range(n_iters):

            batch_reward, batch_loss, batch_num_fail = self.policy_gradient(n_batch, gamma, topo_idx)
            reward += batch_reward
            loss += batch_loss
            num_fail += batch_num_fail

            if n_iter % n_print == 0:
                self.writer.add_scalar("epoch", n_iter, n_iter)
                if n_iter != 0:
                    print(f"########## iteration {n_iter} ##########")
                    print(
                        f"train | reward : {reward / n_print:.3f} ,"
                        f" loss : {loss / n_print:.3f} , "
                        f"num_fail : {num_fail / n_print:.3f}")
                    self.writer.add_scalar("train/loss", loss / n_print, n_iter)
                    self.writer.add_scalar("train/reward", reward / n_print, n_iter)
                    self.writer.add_scalar("train/num_fail", num_fail / n_print, n_iter)
                reward = 0
                loss = 0
                num_fail = 0

                test_reward, test_output, test_num_fail, gen_mofs = self.test_estimate(n_to_generate, topo_idx)

                test_reward_rmsd = np.mean(test_reward['rmsd'])
                test_reward_target = np.mean(test_reward["target"])
                total_test_reward = test_reward_rmsd + test_reward_target

                test_output_rmsd = np.mean(test_output['rmsd'])
                test_output_target = np.mean(test_output["target"])

                print(
                    f"test  | reward_rmsd : {test_reward_rmsd:.3f} , reward_target : {test_reward_target:.3f}, "
                    f"output_rmsd : {test_output_rmsd:.3f}, output_target : {test_output_target:.3f},"
                    f" num_fail : {test_num_fail}/{n_to_generate}")
                print(gen_mofs[:5])

                self.write_logs_test_and_estimate(test_reward,
                                                  test_output,
                                                  test_num_fail,
                                                  gen_mofs,
                                                  n_iter,
                                                  n_to_generate,
                                                  )

                if total_test_reward > callback:
                    path_ckpt = os.path.join(self.log_dir, f"reinforce_model_{n_iter}.ckpt")
                    torch.save(self.generator.state_dict(), path_ckpt)
                    print("model save !!!")
                    callback = total_test_reward



import json
import random
from tqdm import tqdm

import numpy as np
import libs.selfies as sf
from rdkit import Chem

import torch
from pytorch_lightning import LightningModule

from generator import objectives
from generator.transformer import Transformer
from utils import module_utils

from utils.metrics import Metrics
from rdkit.Chem.Draw import MolToImage


class Generator(LightningModule):
    def __init__(self, config):
        super(Generator, self).__init__()
        self.save_hyperparameters()

        self.max_len = config["max_len"]
        # topo
        path_topo_to_idx = config["path_topo_to_idx"]
        self.topo_to_idx = json.load(open(path_topo_to_idx))
        self.idx_to_topo = {v: k for k, v in self.topo_to_idx.items()}
        # mc
        path_mc_to_idx = config["path_mc_to_idx"]
        self.mc_to_idx = json.load(open(path_mc_to_idx))
        self.idx_to_mc = {v: k for k, v in self.mc_to_idx.items()}
        # ol
        path_vocab = config["path_vocab"]
        self.vocab_to_idx = json.load(open(path_vocab))
        self.idx_to_vocab = {v: k for k, v in self.vocab_to_idx.items()}


        self.transformer = Transformer(
            input_dim=len(self.vocab_to_idx),
            output_dim=len(self.vocab_to_idx),
            topo_dim=len(self.topo_to_idx),
            mc_dim=len(self.mc_to_idx),
            hid_dim=config["hid_dim"],
            n_layers=config["n_layers"],
            n_heads=config["n_heads"],
            pf_dim=config["pf_dim"],
            dropout=config["dropout"],
            max_len=config["max_len"],
            src_pad_idx=config["src_pad_idx"],
            trg_pad_idx=config["trg_pad_idx"],
        )

        module_utils.set_metrics(self)
        # ===================== load model ======================

        if config["load_path"] != "":
            ckpt = torch.load(config["load_path"], map_location="cpu")
            state_dict = ckpt["state_dict"]
            self.load_state_dict(state_dict, strict=False)
            print(f"load model : {config['load_path']}")

    def infer(self, batch):
        src = batch["encoded_input"]  # [B, max_len]

        tgt_input = batch["encoded_output"][:, :-1]  # [B, seq_len-1]
        tgt_label = batch["encoded_output"][:, 1:]  # [B, seq_len-1]

        # get mask
        out = self.transformer(src, tgt_input)  # [B, seq_len-1, vocab_dim]

        out.update({
            "src": src,
            "tgt": batch["encoded_output"],
            "tgt_label": tgt_label,
        })

        return out

    def evaluate(self, src, max_len=128):
        """
        src : torch.LongTensor [B=1, seq_len]
        """
        vocab_to_idx = self.vocab_to_idx
        src_mask = self.transformer.make_src_mask(src)

        # get encoded src
        enc_src = self.transformer.encoder(src, src_mask)  # [B=1, seq_len, hid_dim]

        # get target
        trg_indexes = [vocab_to_idx["[SOS]"]]
        for i in range(max_len):
            trg_tensor = torch.LongTensor(trg_indexes).unsqueeze(0).to(src.device)
            trg_mask = self.transformer.make_trg_mask(trg_tensor)  # [B=1, 1, seq_len, seq_len]

            output = self.transformer.decoder(trg_tensor, enc_src, trg_mask, src_mask)  # [B=1, seq_len, vocab_dim]

            if "output_ol" in output.keys():
                out = output["output_ol"]
                pred_token = out.argmax(-1)[:, -1].item()
            elif "output_mc" in output.keys():
                out = output["output_mc"]
                pred_token = out.argmax(-1).item()
            else:
                out = output["output_topo"]
                pred_token = out.argmax(-1).item()
            trg_indexes.append(pred_token)

            if i > 3 and pred_token == vocab_to_idx["[EOS]"]:
                break

        # get topo
        topo_idx = trg_indexes[1]
        topo = self.idx_to_topo[topo_idx]
        # get mc
        mc_idx = trg_indexes[2]
        mc = self.idx_to_mc[mc_idx]
        # get ol
        ol_idx = trg_indexes[3:]
        ol_tokens = [self.idx_to_vocab[idx] for idx in ol_idx]
        # convert to selfies and smiles
        gen_sf = None
        gen_sm = None
        try:
            gen_sf = "".join(ol_tokens[:-1])  # remove EOS token
            gen_sm = sf.decoder(gen_sf)
            m = Chem.MolFromSmiles(gen_sm)
            gen_sm = Chem.MolToSmiles(m) # canonical smiles
        except Exception as e:
            print(e)
            pass

        ret = {
            "topo" : topo,
            "mc" : mc,
            "topo_idx" : topo_idx,
            "mc_idx" : mc_idx,
            "ol_idx" : ol_idx,
            "gen_sf" : gen_sf,
            "gen_sm" : gen_sm,
        }
        return ret

    def forward(self, batch):
        ret = dict()
        ret.update(objectives.compute_loss(self, batch))

        return ret

    def training_step(self, batch, batch_idx):
        ret = self(batch)
        total_loss = sum([v for k, v in ret.items() if "loss" in k])

        return total_loss

    def training_epoch_end(self, outputs):
        module_utils.epoch_wrapup(self)

    def validation_step(self, batch, batch_idx):
        output = self(batch)

    def validation_epoch_end(self, output):
        module_utils.epoch_wrapup(self)

    def test_step(self, batch, batch_idx):
        return batch

    def test_epoch_end(self, batches):
        split = "test"
        module_utils.epoch_wrapup(self)

        metrics = Metrics(self.vocab_to_idx, self.idx_to_vocab)
        list_src = torch.concat([b["encoded_input"] for b in batches], dim=0)

        for src in tqdm(list_src):
            out = self.evaluate(src.unsqueeze(0))

            if out["gen_sm"] is None:
                metrics.num_fail.append(1)
                continue
            else:
                metrics.num_fail.append(0)

            metrics.update(out, src)

        self.log(f"{split}/conn_match", metrics.get_mean(metrics.conn_match))
        self.log(f"{split}/unique_ol", len(set(metrics.gen_ol)))
        self.log(f"{split}/unique_topo_mc", len(set(zip(metrics.gen_topo, metrics.gen_mc))))
        self.log(f"{split}/scaffold", metrics.get_mean(metrics.scaffold))
        self.log(f"{split}/num_fail", metrics.get_mean(metrics.num_fail))


        # add image to log
        # gen_ol with frags (32 images)
        for i in range(32):
            idx = random.Random(i).choice(range(len(metrics.gen_ol)))
            ol = metrics.gen_ol[idx]
            frags = metrics.input_frags[idx]
            imgs = []
            for s in [ol] + frags:
                m = Chem.MolFromSmiles(s)
                if not m:
                    continue
                img = MolToImage(m)
                img = np.array(img)
                img = torch.tensor(img)
                imgs.append(img)
            imgs = np.stack(imgs, axis=0)
            self.logger.experiment.add_image(f"{split}/{i}", imgs, self.global_step, dataformats="NHWC")

        # total gen_ol
        imgs = []
        for i, m in enumerate(metrics.gen_ol[:32]):
            try:
                m = Chem.MolFromSmiles(m)
                img = MolToImage(m)
                img = np.array(img)
                img = torch.tensor(img)
                imgs.append(img)
            except Exception as e:
                print(e)
        imgs = np.stack(imgs, axis=0)
        self.logger.experiment.add_image(f"{split}/gen_ol/", imgs, self.global_step, dataformats="NHWC")

    def configure_optimizers(self):
        return module_utils.set_schedule(self)



import torch
import torch.nn as nn
from pytorch_lightning import LightningModule

from predictor import objectives, heads

from predictor.transformer import Transformer
from utils import module_utils

from torchmetrics.functional import r2_score


class Predictor(LightningModule):
    def __init__(self, config):
        super().__init__()
        self.save_hyperparameters()
        # build transformer
        self.transformer = Transformer(
            embed_dim=config["hid_dim"],
            depth=config["num_layers"],
            num_heads=config["num_heads"],
            mlp_ratio=config["mlp_ratio"],
            drop_rate=config["drop_rate"],
        )

        # metal node embedding
        self.mc_embedding = nn.Embedding(config["mc_dim"], config["hid_dim"])
        self.mc_embedding.apply(module_utils.init_weights)

        # topology embedding
        self.topo_embedding = nn.Embedding(config["topo_dim"], config["hid_dim"])
        self.topo_embedding.apply(module_utils.init_weights)

        # organic linker embedding
        self.ol_embedding = nn.Embedding(config["vocab_dim"], config["hid_dim"], padding_idx=0)
        self.ol_embedding.apply(module_utils.init_weights)

        # class token
        self.cls_embeddings = nn.Linear(1, config["hid_dim"])
        self.cls_embeddings.apply(module_utils.init_weights)

        # position embedding
        # max_len = ol_max_len (100) + cls + mc + topo
        self.pos_embeddings = nn.Parameter(torch.zeros(1, config["max_len"], config["hid_dim"]))
        # self.pos_embeddings.apply(module_utils.init_weights)

        # pooler
        self.pooler = heads.Pooler(config["hid_dim"])
        self.pooler.apply(module_utils.init_weights)

        # ===================== loss =====================

        # regression
        if self.hparams.config["loss_names"]["regression"] > 0:
            self.regression_head = heads.RegressionHead(config["hid_dim"])
            self.regression_head.apply(module_utils.init_weights)
            # normalization
            self.mean = config["mean"]
            self.std = config["std"]
            self.normalizer = module_utils.Normalizer(self.mean, self.std)

        module_utils.set_metrics(self)
        module_utils.set_task(self)
        # ===================== load downstream (test_only) ======================

        if config["load_path"] != "" and config["test_only"]:
            ckpt = torch.load(config["load_path"], map_location="cpu")
            state_dict = ckpt["state_dict"]
            self.load_state_dict(state_dict, strict=False)
            print(f"load model : {config['load_path']}")

    def infer(self, batch):
        mc = batch["mc"]  # [B]
        topo = batch["topo"]  # [B]
        ol = batch["ol"]  # [B, ol_len]
        batch_size = len(mc)

        mc_embeds = self.mc_embedding(mc).unsqueeze(1)  # [B, 1, hid_dim]
        topo_embeds = self.topo_embedding(topo).unsqueeze(1)  # [B, 1, hid_dim]

        ol_embeds = self.ol_embedding(ol)  # [B, ol_len, hid_dim]

        cls_tokens = torch.zeros(batch_size).to(ol_embeds)  # [B]
        cls_embeds = self.cls_embeddings(cls_tokens[:, None, None])  # [B, 1, hid_dim]

        # total_embedding and mask
        co_embeds = torch.cat(
            [cls_embeds, mc_embeds, topo_embeds, ol_embeds],
            dim=1)  # [B, max_len, hid_dim]
        co_masks = torch.cat([torch.ones([batch_size, 3]).to(ol), (ol != 0).float()], dim=1)

        # add pos_embeddings
        final_embeds = co_embeds + self.pos_embeddings
        final_embeds = self.transformer.pos_drop(final_embeds)

        # transformer blocks
        x = final_embeds
        for i, blk in enumerate(self.transformer.blocks):
            x, _attn = blk(x, mask=co_masks)

        x = self.transformer.norm(x)

        cls_feats = self.pooler(x)

        ret = {
            "cls_feats": cls_feats,
            "mc": mc,
            "topo": topo,
            "ol": ol,
            "output": x,
            "output_mask": co_masks,
        }
        return ret

    def forward(self, batch):
        ret = dict()

        if len(self.current_tasks) == 0:
            ret.update(self.infer(batch))

        # regression
        if "regression" in self.current_tasks:
            ret.update(objectives.compute_regression(self, batch, self.normalizer))

        return ret

    def training_step(self, batch, batch_idx):
        module_utils.set_task(self)
        output = self(batch)
        total_loss = sum([v for k, v in output.items() if "loss" in k])

        return total_loss

    def training_epoch_end(self, outputs):
        module_utils.epoch_wrapup(self)

    def validation_step(self, batch, batch_idx):
        module_utils.set_task(self)
        output = self(batch)

    def validation_epoch_end(self, outputs):
        module_utils.epoch_wrapup(self)

    def test_step(self, batch, batch_idx):
        module_utils.set_task(self)
        output = self(batch)
        return output

    def test_epoch_end(self, outputs):
        module_utils.epoch_wrapup(self)

        # calculate r2 score when regression
        if "regression_logits" in outputs[0].keys():
            logits = []
            labels = []
            for out in outputs:
                logits += out["regression_logits"].tolist()
                labels += out["regression_labels"].tolist()
            r2 = r2_score(torch.FloatTensor(logits), torch.FloatTensor(labels))
            self.log(f"test/r2_score", r2)

    def configure_optimizers(self):
        return module_utils.set_schedule(self)



import os
import json
import copy
import torch
from tqdm import tqdm

from utils.metrics import Metrics

from pytorch_lightning import LightningModule
from transformers import get_linear_schedule_with_warmup, get_constant_schedule

from rdkit import Chem
from rdkit.Chem.Draw import MolToImage

import numpy as np
import libs.selfies as sf

topo_to_cn = json.load(open("data/final_topo_cn.json"))
mc_to_cn = json.load(open("data/mc_cn.json"))


class Reinforce(LightningModule):
    def __init__(self, agent, predictors, config):
        super(Reinforce, self).__init__()

        self.agent = copy.deepcopy(agent)
        # freeze only encoder
        for param in self.agent.transformer.encoder.parameters():
            param.requires_grad = False

        self.freeze_agent = copy.deepcopy(agent)
        # freeze agent
        for param in self.freeze_agent.parameters():
            param.requires_grad = False

        self.predictors = predictors

        # reward
        self.threshold = config["threshold"]
        self.reward_max = config["reward_max"]

        # reinforce
        self.lr = config["lr"]
        self.decay_ol = config["decay_ol"]
        self.scheduler = config["scheduler"]
        self.decay_topo = config["decay_topo"]
        self.early_stop = config["early_stop"]
        self.ratio_exploit = config["ratio_exploit"]
        self.ratio_mask_mc = config["ratio_mask_mc"]

        # load model
        if config["load_path"] != "":
            ckpt = torch.load(config["load_path"], map_location="cpu")
            state_dict = ckpt["state_dict"]
            self.load_state_dict(state_dict, strict=False)
            print(f"load model : {config['load_path']}")

    def forward(self, src, max_len=128):
        """
        based on evaluate() in Generator
        :param src: encoded_input of generator [B=1, seq_len]
        :return:
        """

        vocab_to_idx = self.agent.vocab_to_idx

        # mask_mc of encoded_src
        if torch.randn(1) < self.ratio_mask_mc:
            src[:, 0].fill_(0)  # 0 is pad_idx

        src_mask = self.agent.transformer.make_src_mask(src)

        enc_src = self.agent.transformer.encoder(src, src_mask)  # [B=1, seq_len, hid_dim]

        # get target
        trg_indexes = [vocab_to_idx["[SOS]"]]

        list_prob = []

        for i in range(max_len):

            trg_tensor = torch.LongTensor(trg_indexes).unsqueeze(0).to(src.device)
            trg_mask = self.agent.transformer.make_trg_mask(trg_tensor)  # [B=1, 1, seq_len, seq_len]

            output = self.agent.transformer.decoder(
                trg_tensor, enc_src, trg_mask, src_mask)  # [B=1, seq_len, vocab_dim]

            with torch.no_grad():
                freeze_output = self.freeze_agent.transformer.decoder(
                    trg_tensor, enc_src, trg_mask, src_mask)  # [B=1, seq_len, vocab_dim]
                if torch.rand(1)[0] < self.ratio_exploit:
                    output = freeze_output

            if "output_ol" in output.keys():
                out = output["output_ol"]
                prob = out.softmax(dim=-1)[0, -1]
                m = torch.multinomial(prob, 1).item()
            elif "output_mc" in output.keys():
                out = output["output_mc"]
                prob = out.softmax(dim=-1)[0]
                m = torch.multinomial(prob, 1).item()
            else:
                out = output["output_topo"]
                prob = out.softmax(dim=-1)[0]
                m = torch.multinomial(prob, 1).item()

            list_prob.append(prob[m])
            trg_indexes.append(m)

            if i >= 2 and m == vocab_to_idx["[EOS]"]:
                break

        # get topo
        topo_idx = trg_indexes[1]
        topo = self.agent.idx_to_topo[topo_idx]
        # get mc
        mc_idx = trg_indexes[2]
        mc = self.agent.idx_to_mc[mc_idx]
        # get ol
        ol_idx = trg_indexes[3:]
        ol_tokens = [self.agent.idx_to_vocab[idx] for idx in ol_idx]

        # check matching connection points
        topo_cn = topo_to_cn.get(topo, [0])  # if topo is [PAD] then topo_cn = [0]
        if len(topo_cn) == 1:
            topo_cn.append(2)
        mc_cn = mc_to_cn.get(mc, -1)  # if mc is [PAD] then mc_cn = -1
        ol_cn = ol_idx.count(self.agent.vocab_to_idx["[*]"])

        # convert to selfies and smiles
        try:
            gen_sf = "".join(ol_tokens[:-1])  # remove EOS token
            gen_sm = sf.decoder(gen_sf)
            new_gen_sf = sf.encoder(gen_sm)  # to check valid SELFIES
            assert ol_tokens[-1] == "[EOS]", Exception("The last token is not [EOS]")
            assert new_gen_sf == gen_sf, Exception("SELFIES error")
            assert set(topo_cn) == set([mc_cn, ol_cn]), Exception("connection points error")
        except Exception as e:
            print(e)
            # print(f"The failed gen_sf : {gen_sf}")
            # print(f"The failed gen_sm : {gen_sm}")
            gen_sf = None
            gen_sm = None

        ret = {
            "topo": topo,
            "mc": mc,
            "topo_idx": topo_idx,
            "mc_idx": mc_idx,
            "ol_idx": ol_idx,
            "gen_sf": gen_sf,
            "gen_sm": gen_sm,
            "list_prob": list_prob,
        }
        return ret

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=self.lr)
        max_steps = self.trainer.estimated_stepping_batches
        print(f"mas_steps : {max_steps}")

        if self.scheduler == "warmup":
            scheduler = get_linear_schedule_with_warmup(
                optimizer=optimizer,
                num_warmup_steps=int(self.trainer.max_epochs * 0.1),
                num_training_steps=self.trainer.max_epochs,
            )
        elif self.scheduler == "constant":
            scheduler = get_constant_schedule(
                optimizer=optimizer,
            )
        return [optimizer], [scheduler]

    def training_step(self, batch, batch_idx):
        self.agent.train()

        rl_loss = 0
        total_reward = 0
        num_fail = 0

        for enc_input in batch["encoded_input"]:
            src = enc_input.unsqueeze(0)
            out = self(src)
            if out["gen_sm"] is None:
                num_fail += 1
                continue

            # get reward from predictor
            with torch.no_grad():
                rewards, preds = self.get_reward(output_generator=out)
                reward = sum(rewards)

            total_reward += reward

            # Reinforce algorithm
            discounted_reward = reward
            tmp_loss = []

            # topology, metal cluster (the probability of metal cluster is almost zero, because it is used as input)
            for prob in out["list_prob"][:2]:
                rl_loss -= torch.log(prob) * reward * self.decay_topo
                tmp_loss.append(-torch.log(prob).item() * reward)
            # organic linker (discounter_reward is applied only for organic linkers)
            for i, prob in enumerate(out["list_prob"][2:][::-1]):
                rl_loss -= torch.log(prob) * discounted_reward
                discounted_reward = discounted_reward * self.decay_ol

                tmp_loss.append(-torch.log(prob).item() * discounted_reward)
            print(rewards, preds, out["topo"], out["mc"], out["gen_sm"], out["gen_sf"], src[0, 1].item())
            print(rl_loss, tmp_loss[:2], sum(tmp_loss[2:]))

        n_batch = batch["encoded_input"].shape[0]
        rl_loss = rl_loss / n_batch
        total_reward = total_reward / n_batch
        num_fail = torch.tensor(num_fail) / n_batch

        self.log("train/rl_loss", rl_loss)
        self.log("train/reward", total_reward, prog_bar=True)
        self.log("train/num_fail", num_fail, prog_bar=True)

        return rl_loss

    def validation_step(self, batch, batch_idx):
        return batch

    def validation_epoch_end(self, batches):
        list_src = torch.concat([b["encoded_input"] for b in batches], dim=0)
        metrics = Metrics(self.agent.vocab_to_idx, self.agent.idx_to_vocab)
        split = "val"

        metrics = self.update_metrics(list_src, metrics, split)
        if metrics.get_mean(metrics.scaffold) < self.early_stop:
            raise Exception(f"EarlyStopping for scaffold : scaffold accuracy is less than {self.early_stop}")

    def test_step(self, batch, batch_idx):
        return batch

    def test_epoch_end(self, batches):
        list_src = torch.concat([b["encoded_input"] for b in batches], dim=0)
        metrics = Metrics(self.agent.vocab_to_idx, self.agent.idx_to_vocab)
        split = "test"

        metrics = self.update_metrics(list_src, metrics, split)
        # save results
        ret = {
            "rewards": metrics.rewards,
            "preds": metrics.preds,
            "gen_sms": metrics.gen_ol,
            "gen_mcs": metrics.gen_mc,
            "gen_topos": metrics.gen_topo,
        }
        path = os.path.join(self.logger.save_dir, f"results_{self.logger.name}.json")
        json.dump(ret, open(path, "w"))

    def get_reward(self,
                   output_generator,
                   sos_idx=1,
                   pad_idx=0,
                   max_len=128,
                   ):
        # create the input of predictor
        device = self.predictors[0].device
        batch = {}
        batch["mc"] = torch.LongTensor([output_generator["mc_idx"]]).to(device)
        batch["topo"] = torch.LongTensor([output_generator["topo_idx"]]).to(device)
        batch["ol"] = torch.LongTensor([[sos_idx] +  # add sos_idx
                                        output_generator["ol_idx"][:max_len - 4] +
                                        [pad_idx] * (max_len - 4 - len(output_generator["ol_idx"]))]).to(
            device)  # add pad_idx

        preds = []
        rewards = []
        for i, predictor in enumerate(self.predictors):
            infer = predictor.infer(batch)
            p = predictor.regression_head(infer["cls_feats"])
            p = predictor.normalizer.decode(p)
            p = p.detach().item()
            preds.append(p)

            # get reward
            if self.threshold:
                if abs(p) >= abs(self.reward_max[i]):
                    r = 1
                else:
                    r = 0
                rewards.append(r)
            else:
                rewards.append(p / self.reward_max[i])

        if self.threshold:
            if all(rewards):
                rewards = [1] * len(rewards)
            else:
                rewards = [0] * len(rewards)
        return rewards, preds

    def update_metrics(self, list_src, metrics, split, num_log_mols=16):
        """

        :param list_src: list of source
        :param metrics: dictionary of metrics
        :param split: str, split
        :param num_log_mols: number of molecules for log, which will be saved in tensorboard
        :return:
        :rtype:
        """
        for src in tqdm(list_src):
            out = self(src.unsqueeze(0))

            if out["gen_sm"] is None:
                metrics.num_fail.append(1)
                continue
            else:
                metrics.num_fail.append(0)

            with torch.no_grad():
                rewards, preds = self.get_reward(output_generator=out)

            metrics.update(out, src, rewards=rewards, preds=preds)

        self.log(f"{split}/conn_match", metrics.get_mean(metrics.conn_match))
        self.log(f"{split}/unique_ol", len(set(metrics.gen_ol)) / len(metrics.gen_ol))
        self.log(f"{split}/unique_topo_mc", len(set(zip(metrics.gen_topo, metrics.gen_mc))) / len(metrics.gen_topo))
        self.log(f"{split}/scaffold", metrics.get_mean(metrics.scaffold))

        self.log(f"{split}/total_reward", metrics.get_mean(metrics.rewards))
        for i, reward in enumerate(zip(*metrics.rewards)):
            self.log(f"{split}/reward_{i}", metrics.get_mean(reward))
        for i, pred in enumerate(zip(*metrics.preds)):
            self.log(f"{split}/target_{i}", metrics.get_mean(pred))
        self.log(f"{split}/num_fail", metrics.get_mean(metrics.num_fail))

        # add image to log
        for i in range(num_log_mols):
            ol = metrics.gen_ol[i]
            frags = metrics.input_frags[i]
            imgs = []
            for s in [ol] + frags:
                m = Chem.MolFromSmiles(s)
                if not m:
                    continue
                img = MolToImage(m)
                img = np.array(img)
                img = torch.tensor(img)
                imgs.append(img)
            imgs = np.stack(imgs, axis=0)
            self.logger.experiment.add_image(f"{split}/{i}", imgs, self.global_step, dataformats="NHWC")

        # total gen_ol
        imgs = []
        for i, m in enumerate(metrics.gen_ol[:num_log_mols]):
            try:
                m = Chem.MolFromSmiles(m)
                img = MolToImage(m)
                img = np.array(img)
                img = torch.tensor(img)
                imgs.append(img)
            except Exception as e:
                print(e)
        imgs = np.stack(imgs, axis=0)
        self.logger.experiment.add_image(f"{split}/gen_ol/", imgs, self.global_step, dataformats="NHWC")

        return metrics



import torch
import torch.nn as nn
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence

class BaselineModel(nn.Module):
    def __init__(self, vocab_dim, mc_dim, topo_dim, embed_dim, hid_dim):
        super(BaselineModel, self).__init__()
        # model
        # mc
        """
        self.embedding_mc = nn.Embedding(self.mc_dim, self.embed_dim)
        self.rc_mc = nn.Sequential(
            nn.Linear(self.embed_dim, 1),
        )
        """
        self.rc_mc = nn.Linear(mc_dim, 1)
        # topo
        self.embedding_topo = nn.Embedding(topo_dim, embed_dim)
        self.rc_topo = nn.Sequential(
            nn.Linear(embed_dim, 1),
        )
        # ol
        self.embedding_ol = nn.Embedding(vocab_dim, embed_dim)
        self.rnn = nn.RNN(input_size=embed_dim, hidden_size=hid_dim, num_layers=1)
        self.rc_ol = nn.Linear(hid_dim, 1)
        # total
        self.rc_total = nn.Linear(3, 1)
        self.sigmoid = nn.Sigmoid()

    def forward(self, batch):
        mc = batch["mc"]
        topo = batch["topo"]
        ol_pad = batch["ol_pad"]
        ol_len = batch["ol_len"]
        # mc
        # logit_mc = self.embedding_mc(mc)  # [B, embed_dim]
        # logit_mc = self.rc_mc(logit_mc)  # [B, 1]
        logit_mc = self.rc_mc(mc) # [B, 1]

        # topo
        logit_topo = self.embedding_topo(topo)  # [B, embed_dim]
        logit_topo = self.rc_topo(logit_topo)  # [B, 1]
        # ol
        logit_ol = self.embedding_ol(ol_pad)  # [B, pad_len, embed_dim]
        packed_ol = pack_padded_sequence(logit_ol, ol_len, batch_first=True, enforce_sorted=False)
        output_packed, hidden = self.rnn(packed_ol)  # [B, pad_len, hid_dim]
        output_ol, len_ol = pad_packed_sequence(output_packed, batch_first=True)
        logit_ol = self.rc_ol(output_ol[:, -1, :])  # [B, pad_len, hid_dim] ->[B, 1]

        # total
        logit_total = torch.cat([logit_topo, logit_mc, logit_ol], dim=-1)
        logit_total = self.rc_total(logit_total)
        logit_total = self.sigmoid(logit_total)

        return logit_total



from typing import Optional

from pytorch_lightning import LightningDataModule
from torch.utils.data import DataLoader
from generator.dataset import GeneratorDataset


class GeneratorDatamodule(LightningDataModule):
    def __init__(self, _config):
        super(GeneratorDatamodule, self).__init__()
        self.dataset_dir = _config["dataset_dir"]
        self.batch_size = _config["batch_size"]
        self.num_workers = _config["num_workers"]
        self.max_len = _config["max_len"]
        self.path_vocab = _config["path_vocab"]
        self.path_topo_to_idx = _config["path_topo_to_idx"]
        self.path_mc_to_idx = _config["path_mc_to_idx"]

    @property
    def dataset_cls(self):
        return GeneratorDataset

    def set_train_dataset(self):
        self.train_dataset = self.dataset_cls(
            dataset_dir=self.dataset_dir,
            path_vocab=self.path_vocab,
            path_topo_to_idx=self.path_topo_to_idx,
            path_mc_to_idx=self.path_mc_to_idx,
            split="train",
            max_len=self.max_len,
        )

    def set_val_dataset(self):
        self.val_dataset = self.dataset_cls(
            dataset_dir=self.dataset_dir,
            path_vocab=self.path_vocab,
            path_topo_to_idx=self.path_topo_to_idx,
            path_mc_to_idx=self.path_mc_to_idx,
            split="val",
            max_len=self.max_len,
        )

    def set_test_dataset(self):
        self.test_dataset = self.dataset_cls(
            dataset_dir=self.dataset_dir,
            path_vocab=self.path_vocab,
            path_topo_to_idx=self.path_topo_to_idx,
            path_mc_to_idx=self.path_mc_to_idx,
            split="test",
            max_len=self.max_len,
        )

    def setup(self, stage: Optional[str] = None):
        if stage in (None, "fit"):
            self.set_train_dataset()
            self.set_val_dataset()

        if stage in (None, "test"):
            self.set_test_dataset()

        self.collate = self.dataset_cls.collate


    def train_dataloader(self):
        return DataLoader(self.train_dataset,
                          batch_size=self.batch_size,
                          num_workers=self.num_workers,
                          collate_fn=self.collate,
                          shuffle=True, ##
                          pin_memory=True)

    def val_dataloader(self):
        return DataLoader(self.val_dataset,
                          batch_size=self.batch_size,
                          num_workers=self.num_workers,
                          collate_fn=self.collate,
                          shuffle=False, ##
                          pin_memory=True)

    def test_dataloader(self):
        return DataLoader(self.test_dataset,
                          batch_size=self.batch_size,
                          num_workers=self.num_workers,
                          collate_fn=self.collate,
                          shuffle=False, ##
                          pin_memory=True)






import json
import torch
from rdkit import Chem
from rdkit import RDLogger

import libs.selfies as sf

RDLogger.DisableLog('rdApp.*')

topo_to_cn = json.load(open("data/final_topo_cn.json"))
mc_to_cn = json.load(open("data/mc_cn.json"))


class Metrics():
    def __init__(self, vocab_to_idx, idx_to_vocab):
        # generator
        self.num_fail = []
        self.conn_match = []
        self.scaffold = []
        # collect
        self.input_frags = []
        self.gen_ol = []
        self.gen_topo = []
        self.gen_mc = []
        # reinforce
        self.rewards = []
        self.preds = []
        # vocab
        self.topo_to_cn = topo_to_cn
        self.mc_to_cn = mc_to_cn
        self.vocab_to_idx = vocab_to_idx
        self.idx_to_vocab = idx_to_vocab

    def update(self, output, src, rewards=None, preds=None):
        """
        (1) metric for connection matching
        (2) metric for unique organic linker
        (3) metric for unique topology and metal cluster
        (4) metric for scaffold
        (5) (optional) rewards
        (6) (optional) preds

        :param output:
        :return:
        """
        topo_cn = topo_to_cn.get(output["topo"], [0])  # if topo is [PAD] then topo_cn = [0]
        if len(topo_cn) == 1:
            topo_cn.append(2)
        mc_cn = mc_to_cn.get(output["mc"], -1)  # if mc is [PAD] then mc_cn = -1
        ol_cn = output["ol_idx"].count(self.vocab_to_idx["[*]"])

        # (1) metric for connection matching
        if set(topo_cn) == {mc_cn, ol_cn}:
            self.conn_match.append(1)
        else:
            self.conn_match.append(0)

        # (2) metric for unique organic linker
        # (3) metric for unique topology and metal cluster
        gen_sm = output["gen_sm"]
        self.gen_ol.append(gen_sm)
        self.gen_topo.append(output["topo"])
        self.gen_mc.append(output["mc"])

        # (4) metric for scaffold
        # get frags
        frags = [self.idx_to_vocab[idx.item()] for idx in src.squeeze(0)[3:]]
        frags = frags[:frags.index("[EOS]")]
        frags = "".join(frags)

        frags_sm = [sf.decoder(f) for f in frags.split(".")]
        self.input_frags.append(frags_sm)
        # replace * with H
        du, hy = Chem.MolFromSmiles('*'), Chem.MolFromSmiles('[H]')
        try:
            m = Chem.MolFromSmiles(gen_sm)
            check_gen_sm = Chem.ReplaceSubstructs(m, du, hy, replaceAll=True)[0]

            check_ = True
            for sm in frags_sm:
                if not check_gen_sm.HasSubstructMatch(Chem.MolFromSmiles(sm)):
                    check_ = False
            if check_:
                self.scaffold.append(1)
            else:
                self.scaffold.append(0)
        except:
            pass

        # (5) metric for reward
        self.rewards.append(rewards)
        self.preds.append(preds)

    @staticmethod
    def get_mean(list_):
        return torch.mean(torch.Tensor(list_))



import json
import numpy as np
from pathlib import Path
import pandas as pd
import libs.selfies as sf

import torch
from torch.utils.data import Dataset


class GeneratorDataset(Dataset):
    def __init__(self,
                 dataset_dir,
                 path_vocab,
                 path_topo_to_idx,
                 path_mc_to_idx,
                 split,
                 max_len):
        assert split in ["train", "test", "val"]
        # read vocab_to_idx
        self.vocab_to_idx = json.load(open(path_vocab))
        self.topo_to_idx = json.load(open(path_topo_to_idx))
        self.mc_to_idx = json.load(open(path_mc_to_idx))

        # load dataset
        path_data = Path(dataset_dir, f"{split}.csv")
        csv_ = pd.read_csv(path_data)
        print(f"read file : {path_data}, num_data : {len(csv_)}")
        self.topo = np.array(csv_["topo"])
        self.mc = np.array(csv_["mc"])
        self.num_conn = np.array(csv_["num_conn"])
        self.frags = np.array(csv_["frags"])
        self.selfies = np.array(csv_["selfies"])

        self.encoded_input, self.encoded_output = self.encoding(max_len)


    def encoding(self, max_len):
        # making encoded_input
        encoded_input = []
        for i, f in enumerate(self.frags):
            encoded_frags = [self.vocab_to_idx[v] for v in sf.split_selfies(f)]
            encoded = [self.mc_to_idx[self.mc[i]]] + \
                      [self.num_conn[i]] + \
                      [self.vocab_to_idx["[SOS]"]] + \
                      encoded_frags + \
                      [self.vocab_to_idx["[EOS]"]] + \
                      [self.vocab_to_idx["[PAD]"]] * (max_len - 4 - len(encoded_frags))
            encoded_input.append(encoded)

        # making encoded_output
        encoded_output = []
        for i, f in enumerate(self.selfies):
            encoded_selfies = [self.vocab_to_idx[v] for v in sf.split_selfies(f)]
            encoded = [self.vocab_to_idx["[SOS]"]] + \
                      [self.topo_to_idx[self.topo[i]]] + \
                      [self.mc_to_idx[self.mc[i]]] + \
                      encoded_selfies + \
                      [self.vocab_to_idx["[EOS]"]] + \
                      [self.vocab_to_idx["[PAD]"]] * (max_len - 4 - len(encoded_selfies))
            encoded_output.append(encoded)

        return encoded_input, encoded_output


    def __len__(self):
        return len(self.selfies)

    def __getitem__(self, idx):
        ret = dict()

        ret.update(
            {
                "topo" : self.topo[idx],
                "mc" : self.mc[idx],
                "frags" : self.frags[idx],
                "selfies" : self.selfies[idx],
                "encoded_input" : self.encoded_input[idx],
                "encoded_output" : self.encoded_output[idx],
            }
        )
        return ret

    @staticmethod
    def collate(batch):
        keys = set([key for b in batch for key in b.keys()])
        dict_batch = {k: [dic[k] if k in dic else None for dic in batch] for k in keys}
        dict_batch["encoded_input"] = torch.LongTensor(dict_batch["encoded_input"])
        dict_batch["encoded_output"] = torch.LongTensor(dict_batch["encoded_output"])
        return dict_batch




import copy
import os

import pytorch_lightning as pl
from pytorch_lightning.strategies.ddp import DDPStrategy
from pytorch_lightning.loggers import TensorBoardLogger

from generator.config_generator import ex
from generator.datamodule import GeneratorDatamodule
from generator.module import Generator


@ex.automain
def main(_config):
    _config = copy.deepcopy(_config)
    pl.seed_everything(_config["seed"])

    dm = GeneratorDatamodule(_config)

    model = Generator(_config)

    exp_name = f"{_config['exp_name']}"

    os.makedirs(_config["log_dir"], exist_ok=True)
    checkpoint_callback = pl.callbacks.ModelCheckpoint(
        save_top_k=1,
        verbose=True,
        monitor="val/the_metric",
        mode="max",
        save_last=True,
    )

    logger = pl.loggers.TensorBoardLogger(
        _config["log_dir"],
        name=f'{exp_name}_seed{_config["seed"]}_from_{_config["load_path"].split("/")[-1][:-5]}',
    )

    lr_callback = pl.callbacks.LearningRateMonitor(logging_interval="step")
    callbacks = [checkpoint_callback, lr_callback]

    num_devices = _config["num_devices"]
    if isinstance(num_devices, list):
        num_devices = len(num_devices)

    # gradient accumulation
    accumulate_grad_batches = _config["batch_size"] // (
            _config["per_gpu_batchsize"] * num_devices * _config["num_nodes"]
    )

    trainer = pl.Trainer(
        accelerator="gpu",
        devices=num_devices,
        num_nodes=_config["num_nodes"],
        precision=_config["precision"],
        deterministic=True,
        strategy=DDPStrategy(find_unused_parameters=True),
        max_epochs=_config["max_epochs"],
        callbacks=callbacks,
        logger=logger,
        accumulate_grad_batches=accumulate_grad_batches,
        log_every_n_steps=10,
        resume_from_checkpoint=_config["resume_from"],
        val_check_interval=_config["val_check_interval"],
        gradient_clip_val=_config["gradient_clip_val"],
        # profiler="simple",
    )

    if not _config["test_only"]:
        trainer.fit(model, datamodule=dm)
        trainer.test(model, datamodule=dm)
    else:
        trainer.test(model, datamodule=dm)





import torch
from torch import nn

class MultiHeadAttentionLayer(nn.Module):
    def __init__(self, hid_dim, n_heads, dropout):
        super().__init__()

        assert hid_dim % n_heads == 0

        self.hid_dim = hid_dim
        self.n_heads = n_heads
        self.head_dim = hid_dim // n_heads

        self.fc_q = nn.Linear(hid_dim, hid_dim)
        self.fc_k = nn.Linear(hid_dim, hid_dim)
        self.fc_v = nn.Linear(hid_dim, hid_dim)

        self.fc_o = nn.Linear(hid_dim, hid_dim)

        self.dropout = nn.Dropout(dropout)

        self.scale = torch.sqrt(torch.FloatTensor([self.head_dim]))

    def forward(self, query, key, value, mask=None):
        """
        :param query: [B, seq_len, hid_dim]
        :param key: [B, seq_len, hid_dim]
        :param value: [B, seq_len, hid_dim]
        :param mask: [B, 1, 1, src_len] for src_mask, [B, 1, trg_len, trg_len] for trg_mask
        :return: [B, seq_len, hid_dim], [B, n_heads, seq_len, seq_len]
        """
        batch_size = query.shape[0]

        Q = self.fc_q(query) # [batch size, query len, hid dim]
        K = self.fc_k(key) # [batch size, key len, hid dim]
        V = self.fc_v(value) # [batch size, value len, hid dim]

        Q = Q.view(batch_size, -1, self.n_heads, self.head_dim).permute(0, 2, 1, 3)
        K = K.view(batch_size, -1, self.n_heads, self.head_dim).permute(0, 2, 1, 3)
        V = V.view(batch_size, -1, self.n_heads, self.head_dim).permute(0, 2, 1, 3)
        # [batch size, n heads, query len, head dim]
        # [batch size, n heads, key len, head dim]
        # [batch size, n heads, value len, head dim]
        energy = torch.matmul(Q, K.permute(0, 1, 3, 2)) / self.scale.to(query.device)
        # energy = [batch size, n heads, query len, key len]
        if mask is not None:
            energy = energy.masked_fill(mask == 0, -1e10)
        attention = torch.softmax(energy, dim=-1) # [batch size, n heads, query len, key len]
        x = torch.matmul(self.dropout(attention), V) # [batch size, n heads, query len, head dim]
        x = x.permute(0, 2, 1, 3).contiguous() # [batch size, query len, n heads, head dim]
        x = x.view(batch_size, -1, self.hid_dim) # [batch size, query len, hid dim]
        x = self.fc_o(x) # [batch size, query len, hid dim]
        return x, attention


class PositionwiseFeedforwardLayer(nn.Module):
    def __init__(self, hid_dim, pf_dim, dropout):
        super().__init__()

        self.fc_1 = nn.Linear(hid_dim, pf_dim)
        self.fc_2 = nn.Linear(pf_dim, hid_dim)

        self.dropout = nn.Dropout(dropout)

    def forward(self, x):
        """
        :param x: [B, seq_len, hid dim]
        :return: [B, seq_len, hid_dim]
        """
        x = self.dropout(torch.relu(self.fc_1(x))) # [batch size, seq len, pf dim]
        x = self.fc_2(x) # [batch size, seq len, hid dim]
        return x


class EncoderLayer(nn.Module):
    def __init__(self,
                 hid_dim,
                 n_heads,
                 pf_dim,
                 dropout,
                 ):
        super().__init__()

        self.self_attn_layer_norm = nn.LayerNorm(hid_dim)
        self.ff_layer_norm = nn.LayerNorm(hid_dim)
        self.self_attention = MultiHeadAttentionLayer(hid_dim, n_heads, dropout)
        self.positionwise_feedforward = PositionwiseFeedforwardLayer(hid_dim,
                                                                     pf_dim,
                                                                     dropout)
        self.dropout = nn.Dropout(dropout)

    def forward(self, src, src_mask):
        """
        :param src: [B, src_len, hid_dim]
        :param src_mask: [B, 1, 1, src_len]
        :return: [B, src_len, hid_dim]
        """
        # self attention
        _src, _ = self.self_attention(src, src, src, src_mask) # [batch_size, src_len, hid_dim]

        # dropout, residual connection and layer norm
        src = self.self_attn_layer_norm(src + self.dropout(_src)) # [batch_size, src_len, hid_dim]

        # positionwise feedforward
        _src = self.positionwise_feedforward(src) # [batch_size, src_len, hid_dim]

        # dropout, residual and layer norm
        src = self.ff_layer_norm(src + self.dropout(_src)) # [batch_size, src_len, hid_dim]
        return src


class DecoderLayer(nn.Module):
    def __init__(self,
                 hid_dim,
                 n_heads,
                 pf_dim,
                 dropout,
                ):
        super().__init__()

        self.self_attn_layer_norm = nn.LayerNorm(hid_dim)
        self.enc_attn_layer_norm = nn.LayerNorm(hid_dim)
        self.ff_layer_norm = nn.LayerNorm(hid_dim)
        self.self_attention = MultiHeadAttentionLayer(hid_dim, n_heads, dropout)
        self.encoder_attention = MultiHeadAttentionLayer(hid_dim, n_heads, dropout)
        self.positionwise_feedforward = PositionwiseFeedforwardLayer(hid_dim,
                                                                     pf_dim,
                                                                     dropout)
        self.dropout = nn.Dropout(dropout)

    def forward(self, trg, enc_src, trg_mask, src_mask):
        """
        :param trg: [B, trg_len, hid_dim]
        :param enc_src: [B, src_len, hid_dim]
        :param trg_mask: [B, 1, trg_len, trg_len]
        :param src_mask: [B, 1, 1, seq_len]
        :return: [B, trg_len, hid_dim], [B, n_heads, trg_len, src_len]
        """

        # self attention
        _trg, _ = self.self_attention(trg, trg, trg, trg_mask) # [batch size, trg len, hid dim]

        # dropout, residual connection and layer norm
        trg = self.self_attn_layer_norm(trg + self.dropout(_trg)) # [batch size, trg len, hid dim]

        # encoder attention
        _trg, attention = self.encoder_attention(trg, enc_src, enc_src, src_mask)
        # [batch size, trg len, hid dim], [batch size, n heads, trg len, src len]

        # dropout, residual connection and layer norm
        trg = self.enc_attn_layer_norm(trg + self.dropout(_trg)) # [batch size, trg len, hid dim]

        # positionwise feedforward
        _trg = self.positionwise_feedforward(trg) # [batch size, trg len, hid dim]

        # dropout, residual and layer norm
        trg = self.ff_layer_norm(trg + self.dropout(_trg)) # [batch size, trg len, hid dim]

        return trg, attention


import json
from sacred import Experiment

ex = Experiment("predictor")

mc_to_idx = json.load(open("data/mc_to_idx.json"))
topo_to_idx = json.load(open("data/topo_to_idx.json"))
vocab_to_idx = json.load(open("data/vocab_to_idx.json"))  # vocab for selfies

def _loss_names(d):
    ret = {
        "regression": 0,  # regression
    }
    ret.update(d)
    return ret

@ex.config
def config():
    seed = 0
    exp_name = "predictor"

    dataset_dir = "###"
    loss_names = _loss_names({"regression": 1})

    # model setting
    max_len = 128 # cls + mc + topo + ol_len
    vocab_dim = len(vocab_to_idx)
    mc_dim = len(mc_to_idx)
    topo_dim = len(topo_to_idx)

    # transformer setting
    hid_dim = 256
    num_heads = 4  # hid_dim / 64
    num_layers = 4
    mlp_ratio = 4
    drop_rate = 0.1

    # run setting
    batch_size = 128
    per_gpu_batchsize = 64
    load_path = ""
    log_dir = "predictor/logs"
    num_workers = 8  # recommend num_gpus * 4
    num_nodes = 1
    num_gpus = 2
    precision = 16

    # downstream
    downstream = ""
    n_classes = 0
    threshold_classification = None

    # Optimizer Setting
    optim_type = "adamw"  # adamw, adam, sgd (momentum=0.9)
    learning_rate = 1e-4
    weight_decay = 1e-2
    decay_power = 1  # default polynomial decay, [cosine, constant, constant_with_warmup]
    max_epochs = 50
    max_steps = -1  # num_data * max_epoch // batch_size (accumulate_grad_batches)
    warmup_steps = 0.05  # int or float ( max_steps * warmup_steps)
    end_lr = 0
    lr_mult = 1  # multiply lr for downstream heads

    # trainer setting
    resume_from = None
    val_check_interval = 1.0
    test_only = False

    # normalize (when regression)
    mean = None
    std = None

@ex.named_config
def env_ifactor():
    pass

@ex.named_config
def regression_vf():
    exp_name = "regression_vf"
    dataset_dir = "data/dataset_predictor/vf"

    # trainer
    max_epochs = 50
    batch_size = 128
    per_gpu_batchsize = 16


@ex.named_config
def regression_qkh():
    exp_name = "regression_qkh"
    dataset_dir = "data/dataset_predictor/qkh"

    # trainer
    max_epochs = 50
    batch_size = 64
    per_gpu_batchsize = 16

    # normalize (when regression)
    mean = -19.408
    std = -9.172


@ex.named_config
def regression_selectivity():
    exp_name = "regression_selectivity"
    dataset_dir = "data/dataset_predictor/selectivity"

    # trainer
    max_epochs = 50
    batch_size = 128
    per_gpu_batchsize = 16

    # normalize (when regression)
    mean = 1.872
    std = 1.922


"""
Round 2
"""
@ex.named_config
def regression_qkh_round2():
    exp_name = "regression_qkh_round2"
    dataset_dir = "data/dataset_predictor/qkh/round2/"

    # trainer
    max_epochs = 50
    batch_size = 64
    per_gpu_batchsize = 16

    # normalize (when regression)
    mean = -19.886
    std = 9.811


import math
import torch
import torch.nn as nn


def attention(query, key, value, mask=None, dropout=None):
    """

    :param query: [B, num_heads, seq_len, hid_dim//num_heads]
    :param key: [B, num_heads, seq_len, hid_dim//num_heads]
    :param value: [B, num_heads, seq_len, hid_dim//num_heads]
    :param mask: [B, 1, 1, seq_len]
    :param dropout: (float) dropout_rate
    :return:  [B, num_heads, seq_len, hid_dim//num_heads]
    """
    "Compute 'Scaled Dot Product Attention'"
    d_k = query.size(-1) # hid_dim
    scores = torch.matmul(query, key.transpose(-2, -1)) / math.sqrt(d_k)  # [B, num_heads, seq_len, seq_len]
    if mask is not None:
        scores = scores.masked_fill(mask == 0, -float('inf'))  # [B, num_heads, seq_len, seq_len]

    p_attn = scores.softmax(dim=-1)  # [B, num_heads, seq_len, seq_len]

    if dropout is not None:
        p_attn = dropout(p_attn)
    return torch.matmul(p_attn, value), p_attn  # [B, num_heads, seq_len, hid_dim//num_heads]


class MultiHeadedAttention(nn.Module):
    def __init__(self, h, d_model, dropout_rate=0.1):
        super(MultiHeadedAttention, self).__init__()
        assert d_model % h == 0
        # We assume d_v always equals d_k
        self.d_k = d_model // h
        self.h = h
        self.layer_q = nn.Linear(d_model, d_model)
        self.layer_k = nn.Linear(d_model, d_model)
        self.layer_v = nn.Linear(d_model, d_model)
        self.proj = nn.Linear(d_model, d_model)

        self.dropout = nn.Dropout(p=dropout_rate)

    def forward(self, q, k, v, mask=None):
        # query : [B, seq_len, hid_dim]
        # mask : [B, 1, seq_len] for src, [B, seq_len, seq_len] for tgt
        if mask is not None:
            mask = mask.unsqueeze(1)  # [B, 1, 1, seq_len]
        batch_size = q.size(0)

        # [B, seq_len, num_heads, hid_dim//num_heads] -> [B, num_heads, seq_len, hid_dim//num_heads]
        q = self.layer_q(q).view(batch_size, -1, self.h, self.d_k).transpose(1, 2)
        k = self.layer_k(k).view(batch_size, -1, self.h, self.d_k).transpose(1, 2)
        v = self.layer_k(v).view(batch_size, -1, self.h, self.d_k).transpose(1, 2)

        # 2) Apply attention on all the projected vectors in batch.
        x, p_attn = attention(
            q, k, v, mask=mask, dropout=self.dropout
        )  # [B, num_heads, seq_len, hid_dim//num_heads]

        # 3) "Concat" using a view and apply a final linear.
        x = (
            x.transpose(1, 2)  # [B, seq_len, num_heads, hid_dim//num_heads]
                .contiguous()
                .view(batch_size, -1, self.h * self.d_k)  # [B, seq_len, hid_dim]
        )

        x = self.proj(x)
        return x, p_attn



import copy
import os

import pytorch_lightning as pl
from pytorch_lightning.strategies.ddp import DDPStrategy
from pytorch_lightning.loggers import TensorBoardLogger

from reinforce.config_reinforce import ex

from predictor.config_predictor import config as predictor_config
from predictor.config_predictor import _loss_names
from predictor.module import Predictor

from generator.config_generator import config as generator_config
from generator.datamodule import GeneratorDatamodule
from generator.module import Generator

from reinforce.module import Reinforce

@ex.automain
def main(_config):
    # 1. load predictor
    predictors = []
    for i in range(len(_config["predictor_load_path"])):
        pred_config = predictor_config()
        pred_config["test_only"] = True
        pred_config["loss_names"] = _loss_names({"regression": 1})
        pred_config["load_path"] = _config["predictor_load_path"][i]
        pred_config["mean"] = _config["mean"][i]
        pred_config["std"] = _config["std"][i]

        predictor = Predictor(pred_config)
        predictor.eval()
        predictors.append(predictor)

    # 2. load generator
    gen_config = generator_config()
    gen_config["load_path"] = _config["generator_load_path"]
    gen_config["batch_size"] = _config["batch_size"]
    gen_config["dataset_dir"] = _config["dataset_dir"]

    generator = Generator(gen_config)
    dm = GeneratorDatamodule(gen_config)

    # 3. set reinforce
    _config = copy.deepcopy(_config)
    pl.seed_everything(_config["seed"])

    model = Reinforce(generator, predictors, _config)

    exp_name = f"{_config['exp_name']}"
    os.makedirs(_config["log_dir"], exist_ok=True)
    checkpoint_callback = pl.callbacks.ModelCheckpoint(
        save_top_k=1,
        verbose=True,
        monitor="val/total_reward",
        mode="max",
        save_last=True,
    )

    logger = pl.loggers.TensorBoardLogger(

        _config["log_dir"],
        name=f'{exp_name}_seed{_config["seed"]}_from_{_config["load_path"].split("/")[-1][:-5]}',
    )
    lr_callback = pl.callbacks.LearningRateMonitor(logging_interval="step")
    callbacks = [checkpoint_callback, lr_callback]

    trainer = pl.Trainer(
        accelerator="gpu",
        devices=_config["devices"],
        num_nodes=_config["num_nodes"],
        precision=_config["precision"],
        deterministic=True,
        strategy=DDPStrategy(find_unused_parameters=True),
        max_epochs=_config["max_epochs"],
        callbacks=callbacks,
        logger=logger,
        accumulate_grad_batches=_config["accumulate_grad_batches"],
        log_every_n_steps=10,
        resume_from_checkpoint=_config["resume_from"],
        limit_train_batches=_config["limit_train_batches"],
        limit_val_batches=_config["limit_val_batches"],
        num_sanity_val_steps=_config["limit_val_batches"],
        val_check_interval=_config["val_check_interval"],
        gradient_clip_val=_config["gradient_clip_val"],
    )

    if not _config["test_only"]:
        trainer.fit(model, datamodule=dm)
        trainer.test(model, datamodule=dm)
    else:
        trainer.test(model, datamodule=dm)





import re
from setuptools import setup, find_packages

try:
    import torch
except ImportError:
    raise EnvironmentError('Torch must be installed before installation')

with open("requirements.txt", "r") as f:
    install_requires = f.readlines()

with open("README.md", "r") as f:
    long_description=f.read()

extras_require = {
    'docs': ['sphinx', 'livereload', 'myst-parser']
}

# with open('mofreinforce/__init__.py') as f:
#     version = re.search(r"__version__ = '(?P<version>.+)'", f.read()).group('version')


setup(
    name='mofreinforce',
    version="0.0.2",
    description='mofreinforce',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Hyunsoo Park',
    author_email='phs68660888@gmail.com',
    packages=find_packages(),
    package_data={'mofreinforce': []},
    install_requires=install_requires,
    extras_require=extras_require,
    scripts=[],
    download_url='https://github.com/hspark1212/MOFreinforce',
    entry_points={'console_scripts':['mofreinforce=mofreinforce.cli.main:main']},
    python_requires='>=3.8',
)



import torch.nn.functional as F
from torchmetrics.functional import r2_score


def compute_regression(pl_module, batch, normalizer):
    infer = pl_module.infer(batch)
    batch_size = pl_module.hparams.config["batch_size"]

    logits = pl_module.regression_head(infer["cls_feats"]).squeeze(-1)  # [[B]]
    target = batch["target"]  # [B]

    # normalize encode if config["mean"] and config["std], else pass
    target = normalizer.encode(target)

    loss = F.mse_loss(logits, target)
    ret = {
        "regression_loss": loss,
        "regression_logits": normalizer.decode(logits),
        "regression_labels": normalizer.decode(target),
    }

    # call update() loss and acc
    phase = "train" if pl_module.training else "val"
    loss = getattr(pl_module, f"{phase}_regression_loss")(ret["regression_loss"])
    mae = getattr(pl_module, f"{phase}_regression_mae")(
        F.l1_loss(ret["regression_logits"], ret["regression_labels"])
    )

    r2 = getattr(pl_module, f"{phase}_regression_r2")(
        r2_score(logits, target)
    )

    pl_module.log(f"regression/{phase}/loss", loss, batch_size=batch_size, prog_bar=True)
    pl_module.log(f"regression/{phase}/mae", mae, batch_size=batch_size, prog_bar=True)
    pl_module.log(f"regression/{phase}/r2", r2, batch_size=batch_size, prog_bar=True)
    return ret



import torch
from torchmetrics import Metric


class Accuracy(Metric):
    def __init__(self, dist_sync_on_step=False):
        super().__init__(dist_sync_on_step=dist_sync_on_step)
        self.add_state("correct", default=torch.tensor(0.0), dist_reduce_fx="sum")
        self.add_state("total", default=torch.tensor(0.0), dist_reduce_fx="sum")

    def update(self, logits, target):
        logits, target = (
            logits.detach().to(self.correct.device),
            target.detach().to(self.correct.device),
        )
        if len(logits.shape) > 1:
            preds = logits.argmax(dim=-1)
        else:
            # binary accuracy
            logits[logits >= 0.5] = 1
            logits[logits < 0.5] = 0
            preds = logits

        preds = preds[target != -100]
        target = target[target != -100]

        if target.numel() == 0:
            return 1

        assert preds.shape == target.shape

        self.correct += torch.sum(preds == target)
        self.total += target.numel()

    def compute(self):
        return self.correct / self.total


class Scalar(Metric):
    def __init__(self, dist_sync_on_step=False):
        super().__init__(dist_sync_on_step=dist_sync_on_step)
        self.add_state("scalar", default=torch.tensor(0.0), dist_reduce_fx="sum")
        self.add_state("total", default=torch.tensor(0.0), dist_reduce_fx="sum")

    def update(self, scalar):
        if isinstance(scalar, torch.Tensor):
            scalar = scalar.detach().to(self.scalar.device)
        else:
            scalar = torch.tensor(scalar).float().to(self.scalar.device)
        self.scalar += scalar
        self.total += 1

    def compute(self):
        return self.scalar / self.total